# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2006-2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2008 Henry Obein <henry@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from datetime import datetime
from operator import itemgetter

# Import from itools
from itools.datatypes import DateTime, String
from itools.gettext import MSG
from itools.i18n import format_datetime
from itools.web import get_context, STLView
from itools.xapian import KeywordField, BoolField

# Import from ikaaro
from metadata import Record
from resource_ import DBResource


###########################################################################
# Views
###########################################################################
class HistoryView(STLView):

    access = 'is_allowed_to_view'
    title = MSG(u'History')
    icon = 'history.png'
    template = '/ui/file/history.xml'


    def get_namespace(self, resource, context):
        return {
            'revisions': resource.get_revisions(context),
        }



###########################################################################
# Model
###########################################################################
class History(Record):

    schema = {
        'date': DateTime,
        'user': String,
        'size': String}


class VersioningAware(DBResource):

    @classmethod
    def get_metadata_schema(cls):
        schema = DBResource.get_metadata_schema()
        schema['history'] = History
        return schema


    ########################################################################
    # API
    ########################################################################
    def commit_revision(self):
        context = get_context()
        username = ''
        if context is not None:
            user = context.user
            if user is not None:
                username = user.name

        property = {'user': username,
                    'date': datetime.now(),
                    'size': str(self.get_size())}
        self.metadata.set_property('history', property)


    def get_revisions(self, context):
        accept = context.accept_language
        revisions = []

        for revision in self.get_property('history'):
            date = revision['date']
            revisions.append({
                'username': revision['user'],
                'date': format_datetime(date, accept=accept),
                'sort_date': date,
                'size': revision['size']})

        revisions.sort(key=itemgetter('sort_date'), reverse=True)
        return revisions


    def get_owner(self):
        history = self.get_property('history')
        if not history:
            return None
        return history[0]['user']


    def get_last_author(self):
        history = self.get_property('history')
        if history:
            return history[-1]['user']


    def get_mtime(self):
        history = self.get_property('history')
        if not history:
            return DBResource.get_mtime(self)
        return history[-1]['date']


    ########################################################################
    # Index & Search
    ########################################################################
    def get_catalog_fields(self):
        return DBResource.get_catalog_fields(self) + [
            # Versioning Aware
            BoolField('is_version_aware'),
            KeywordField('last_author', is_indexed=False, is_stored=True)]


    def get_catalog_values(self):
        document = DBResource.get_catalog_values(self)

        document['is_version_aware'] = True
        # Last Author (used in the Last Changes view)
        last_author = self.get_last_author()
        if last_author is not None:
            users = self.get_resource('/users')
            try:
                user = users.get_resource(last_author)
            except LookupError:
                document['last_author'] = None
            else:
                document['last_author'] = user.get_title()

        return document


    ########################################################################
    # User Interface
    ########################################################################
    history = HistoryView()



# -*- coding: UTF-8 -*-
# Copyright (C) 2008 David Versmisse <david.versmisse@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from operator import itemgetter

# Import from itools
from itools.datatypes import Enumerate, String, Integer, Boolean
from itools.web import get_context



class TrackerList(Enumerate):

    @staticmethod
    def decode(value):
        if not value:
            return None
        return int(value)


    @staticmethod
    def encode(value):
        if value is None:
            return ''
        return str(value)


    @classmethod
    def get_options(cls):
        elements = cls.tracker.get_resource(cls.element).handler
        return [{'name': record.id,
                 'value': elements.get_record_value(record, 'title')}
                for record in elements.get_records_in_order()]



class ProductInfoList(Enumerate):

    @staticmethod
    def decode(value):
        if not value:
            return None
        return int(value)


    @staticmethod
    def encode(value):
        if value is None:
            return ''
        return str(value)


    @classmethod
    def get_options(cls):
        tracker = cls.tracker
        products = tracker.get_resource('product').handler
        elements = tracker.get_resource(cls.element).handler

        options = []
        for record in elements.get_records_in_order():
            title = elements.get_record_value(record, 'title')
            product_id = elements.get_record_value(record, 'product')

            # Product title
            if product_id is None:
                continue
            product_id = int(product_id)
            product_record = products.get_record(product_id)
            product_title = products.get_record_value(product_record, 'title')

            options.append({'name': record.id,
                            'value': '%s - %s' % (product_title, title)})
        return options


    @classmethod
    def is_valid(cls, name):
        # Get the product number
        product =  get_context().get_form_value('product')
        if product is None:
            return True
        product = int(product)

        # Match our choice ?
        choice = int(name)
        elements = cls.tracker.get_resource(cls.element).handler
        record = elements.get_record(choice)
        product_id = int(elements.get_record_value(record, 'product'))

        return product_id == product



class UsersList(Enumerate):
    @classmethod
    def get_options(cls):
        site_root = cls.tracker.get_site_root()
        users = site_root.get_resource('/users')
        options = [{'name': x,
                    'value': users.get_resource(x).get_title()}
                    for x in site_root.get_members()]
        options.sort(key=itemgetter('value'))
        return options


def get_issue_fields(resource_tracker):
    return {'title': String(mandatory=True),
            'product': TrackerList(element='product',
                                   tracker=resource_tracker, mandatory=True),
            'module': ProductInfoList(element='module',
                                      tracker=resource_tracker),
            'version': ProductInfoList(element='version',
                                       tracker=resource_tracker),
            'type': TrackerList(element='type', tracker=resource_tracker,
                                mandatory=True),
            'state': TrackerList(element='state', tracker=resource_tracker,
                                 mandatory=True),
            'priority': TrackerList(element='priority',
                                    tracker=resource_tracker),
            'assigned_to': UsersList(tracker=resource_tracker),
            'cc_list': UsersList(tracker=resource_tracker, multiple=True),
            'comment': String,
            'file': String}









# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007 Luis Arturo Belmar-Letelier <luis@itaapy.com>
# Copyright (C) 2007 Nicolas Deram <nicolas@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from itools
from itools.datatypes import Boolean, Integer, String, Unicode
from itools.datatypes import Enumerate
from itools.gettext import MSG
from itools.handlers import merge_dicts
from itools.xapian import PhraseQuery, AndQuery
from itools.web import ERROR, INFO

# Import from ikaaro
from ikaaro.forms import title_widget, BooleanCheckBox, SelectWidget
from ikaaro.registry import register_resource_class
from ikaaro.table import OrderedTable, OrderedTableFile
from ikaaro.table_views import OrderedTable_View



class ProductsEnumerate(Enumerate):

    @classmethod
    def get_options(cls):
        products = cls.products.handler
        return [
            {'name': str(x.id),
             'value': products.get_record_value(x, 'title')}
            for x in products.get_records_in_order() ]


###########################################################################
# Views
###########################################################################
class SelectTable_View(OrderedTable_View):

    def get_table_columns(self, resource, context):
        cls = OrderedTable_View
        columns = cls.get_table_columns(self, resource, context)
        columns.append(('issues', MSG(u'Issues')))
        return columns


    def get_item_value(self, resource, context, item, column):
        # Append a column with the number of issues
        if column == 'issues':
            filter = str(resource.name)

            # Build the search query
            query_terms = resource.parent.get_issues_query_terms()
            query_terms.append(PhraseQuery(filter, item.id))
            query = AndQuery(*query_terms)

            # Search
            results = context.root.search(query)
            count = len(results)
            if count == 0:
                return 0, None
            return count, '../;view?%s=%s' % (filter, item.id)

        # Default
        cls = OrderedTable_View
        value = cls.get_item_value(self, resource, context, item, column)

        # NOTE The field 'product' is reserved to make a reference to the
        # 'products' table.  Currently it is used by the 'versions' and
        # 'modules' tables.
        if column == 'product':
            value = int(value)
            handler = resource.parent.get_resource('product').handler
            record = handler.get_record(value)
            return handler.get_record_value(record, 'title')

        return value


    def sort_and_batch(self, resource, context, items):
        # Sort
        sort_by = context.query['sort_by']
        if sort_by != 'issues':
            cls = OrderedTable_View
            return cls.sort_and_batch(self, resource, context, items)

        reverse = context.query['reverse']
        f = lambda x: self.get_item_value(resource, context, x, 'issues')[0]
        items.sort(cmp=lambda x,y: cmp(f(x), f(y)), reverse=reverse)

        # Batch
        start = context.query['batch_start']
        size = context.query['batch_size']
        return items[start:start+size]



###########################################################################
# Resources
###########################################################################
class Tracker_TableHandler(OrderedTableFile):

    record_schema = {'title': Unicode(multiple=True)}



class Tracker_TableResource(OrderedTable):

    class_id = 'tracker_select_table'
    class_title = MSG(u'Select Table')
    class_handler = Tracker_TableHandler

    form = [title_widget]


    def get_options(self, value=None, sort=None):
        # Find out the options
        handler = self.handler
        options = []
        for id in handler.get_record_ids_in_order():
            record = handler.get_record(id)
            title = handler.get_record_value(record, 'title')
            options.append({'id': id, 'title': title})

        # Sort
        if sort is not None:
            options.sort(key=lambda x: x.get(sort))

        # Set 'is_selected'
        if value is None:
            for option in options:
                option['is_selected'] = False
        elif isinstance(value, list):
            for option in options:
                option['is_selected'] = (option['id'] in value)
        else:
            for option in options:
                option['is_selected'] = (option['id'] == value)

        return options


    view = SelectTable_View()


    def del_record_action(self, context):
        # Check input
        ids = context.get_form_values('ids', type=Integer)
        if not ids:
            return context.come_back(ERROR(u'No resource selected.'))

        # Search all issues
        query_terms = resource.parent.get_issues_query_terms()
        query = AndQuery(*query_terms)
        results = context.root.search(query)

        # Remove values only if no issues have them
        filter = str(self.name)
        removed = []
        for id in ids:
            query = PhraseQuery(filter, id)
            count = results.search(query).get_n_documents()
            if count == 0:
                self.handler.del_record(id)
                removed.append(str(id))

        # Ok
        message = INFO(u'Resources removed: $resources.')
        return context.come_back(message, resources=', '.join(removed))



class ModulesHandler(Tracker_TableHandler):

    record_schema = {
        'product': String(mandatory=True),
        'title': Unicode(multiple=True, mandatory=True)}



class ModulesResource(Tracker_TableResource):

    class_id = 'tracker_modules'
    class_handler = ModulesHandler

    def get_schema(self):
        products = self.parent.get_resource('product')
        return merge_dicts(
            ModulesHandler.record_schema,
            product=ProductsEnumerate(products=products, mandatory=True))


    form = [
        SelectWidget('product', title=MSG(u'Product')),
        title_widget]



class VersionsHandler(Tracker_TableHandler):

    record_schema = {
        'product': String(mandatory=True),
        'title': Unicode(multiple=True, mandatory=True),
        'released': Boolean}



class VersionsResource(Tracker_TableResource):

    class_id = 'tracker_versions'
    class_handler = VersionsHandler

    def get_schema(self):
        products = self.parent.get_resource('product')
        return merge_dicts(
            VersionsHandler.record_schema,
            product=ProductsEnumerate(products=products, mandatory=True))


    form = [
        SelectWidget('product', title=MSG(u'Product')),
        title_widget,
        BooleanCheckBox('released', title=MSG(u'Released'))]



###########################################################################
# Register
###########################################################################
register_resource_class(Tracker_TableResource)
register_resource_class(ModulesResource)
register_resource_class(VersionsResource)



# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007-2008 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2008 Nicolas Deram <nicolas@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from decimal import Decimal
from types import GeneratorType

# Import from itools
from itools.datatypes import Tokens
from itools.gettext import MSG
from itools.html import stream_to_str_as_html, xhtml_doctype
from itools.web import STLView
from itools.xml import XMLParser

# Import from ikaaro
from access import RoleAware
from control_panel import ControlPanel, CPAddUser, CPBrokenLinks
from control_panel import CPBrowseUsers, CPEditContactOptions, CPEditLanguages
from control_panel import CPEditMembership, CPEditSecurityPolicy
from control_panel import CPEditVirtualHosts, CPOrphans
from folder import Folder
from registry import register_resource_class
from registry import register_website
from resource_views import LoginView
from skins import UI, ui_path
from website_views import AboutView, ContactForm, CreditsView
from website_views import ForgottenPasswordForm, NewWebSiteForm, RegisterForm
from website_views import SiteSearchView



class WebSite(RoleAware, Folder):

    class_id = 'WebSite'
    class_version = '20071215'
    class_title = MSG(u'Web Site')
    class_description = MSG(u'Create a new Web Site or Work Place.')
    class_icon16 = 'icons/16x16/website.png'
    class_icon48 = 'icons/48x48/website.png'
    class_skin = 'ui/aruni'
    class_views = Folder.class_views + ['control_panel', 'last_changes']
    class_control_panel = ['browse_users', 'add_user', 'edit_virtual_hosts',
                           'edit_security_policy', 'edit_languages',
                           'edit_contact_options', 'broken_links', 'orphans']


    __fixed_handlers__ = ['skin', 'index']


    def _get_resource(self, name):
        if name == 'ui':
            ui = UI(ui_path)
            ui.database = self.metadata.database
            return ui
        if name in ('users', 'users.metadata'):
            return self.parent._get_resource(name)
        return Folder._get_resource(self, name)


    @classmethod
    def get_metadata_schema(cls):
        schema = Folder.get_metadata_schema()
        schema.update(RoleAware.get_metadata_schema())
        schema['vhosts'] = Tokens(default=())
        schema['contacts'] = Tokens(default=())
        schema['website_languages'] = Tokens(default=('en',))
        return schema


    ########################################################################
    # Publish
    ########################################################################
    unauthorized = LoginView()


    ########################################################################
    # API
    ########################################################################
    def get_default_language(self):
        return self.get_property('website_languages')[0]


    def before_traverse(self, context, min=Decimal('0.000001'),
                        zero=Decimal('0.0')):
        # Set the language cookie if specified by the query.
        # NOTE We do it this way, instead of through a specific action,
        # to avoid redirections.
        language = context.get_form_value('language')
        if language is not None:
            context.set_cookie('language', language)

        # The default language (give a minimum weight)
        accept = context.accept_language
        default = self.get_default_language()
        if accept.get(default, zero) < min:
            accept.set(default, min)
        # User Profile (2.0)
        user = context.user
        if user is not None:
            language = user.get_property('user_language')
            if language is not None:
                accept.set(language, 2.0)
        # Cookie (2.5)
        language = context.get_cookie('language')
        if language is not None:
            accept.set(language, 2.5)


    def after_traverse(self, context):
        body = context.entity
        if not isinstance(body, (str, list, GeneratorType, XMLParser)):
            return

        # If there is not content type and the body is not None, wrap it in
        # the skin template
        if context.response.has_header('Content-Type'):
            if isinstance(body, (list, GeneratorType, XMLParser)):
                context.entity = stream_to_str_as_html(body)
            return

        if isinstance(body, str):
            body = XMLParser(body, doctype=xhtml_doctype)
        context.entity = context.root.get_skin().template(body)


    def is_allowed_to_register(self, user, resource):
        return self.get_property('website_is_open')


    #######################################################################
    # UI
    #######################################################################
    new_instance = NewWebSiteForm()
    # Control Panel
    control_panel = ControlPanel()
    browse_users = CPBrowseUsers()
    add_user = CPAddUser()
    edit_membership = CPEditMembership()
    edit_virtual_hosts = CPEditVirtualHosts()
    edit_security_policy = CPEditSecurityPolicy()
    edit_contact_options = CPEditContactOptions()
    edit_languages = CPEditLanguages()
    broken_links = CPBrokenLinks()
    orphans = CPOrphans()
    # Register / Login
    register = RegisterForm()
    forgotten_password = ForgottenPasswordForm()
    # Public views
    site_search = SiteSearchView()
    contact = ContactForm()
    about = AboutView()
    credits = CreditsView()
    license = STLView(access=True, title=MSG(u'License'),
                      template='/ui/root/license.xml')


###########################################################################
# Register
###########################################################################
register_resource_class(WebSite)
register_website(WebSite)



# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2006-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Henry Obein <henry@itaapy.com>
# Copyright (C) 2008 Nicolas Deram <nicolas@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from datetime import datetime

# Import from itools
from itools.datatypes import String
from itools.gettext import MSG
from itools.handlers import File as FileHandler, Image as ImageHandler
from itools.handlers import TARFile, ZIPFile, GzipFile, Bzip2File
from itools.odf import SXWFile, SXCFile, SXIFile, ODTFile, ODSFile, ODPFile
from itools.pdf import PDFFile
from itools.utils import guess_all_extensions
from itools.vfs import FileName
from itools.web import STLView
from itools.xml import MSPowerPoint as MSPowerPointFile, RTF as RTFFile
from itools.xml import MSWord as MSWordFile, MSExcel as MSExcelFile

# Import from ikaaro
from registry import register_resource_class
from versioning import VersioningAware
from workflow import WorkflowAware
from file_views import File_NewInstance, File_Download, File_View
from file_views import File_Edit, File_ExternalEdit, File_Backlinks
from file_views import Image_Thumbnail, Image_View, Video_View, Archive_View
from file_views import Flash_View



###########################################################################
# Base File
###########################################################################
class File(WorkflowAware, VersioningAware):

    class_id = 'file'
    class_version = '20071216'
    class_title = MSG(u'File')
    class_description = MSG(
        u'Upload office documents, images, media files, etc.')
    class_icon16 = 'icons/16x16/file.png'
    class_icon48 = 'icons/48x48/file.png'
    class_views = ['view', 'edit', 'externaledit', 'upload', 'backlinks',
                   'edit_state', 'history']
    class_handler = FileHandler


    @staticmethod
    def _make_resource(cls, folder, name, body=None, filename=None,
                     extension=None, **kw):
        VersioningAware._make_resource(cls, folder, name, filename=filename,
                                       **kw)
        # Add the body
        if body is not None:
            handler = cls.class_handler(string=body)
            if extension:
                extension = extension.lower()
            else:
                extension = handler.class_extension
            name = FileName.encode((name, extension, None))
            folder.set_handler(name, handler)


    def get_all_extensions(self):
        format = self.metadata.format
        # FIXME This is a hack, compression encodings are not yet properly
        # supported (to do for the next major version).
        if format == 'application/x-gzip':
            extensions = ['gz', 'tgz']
        elif format == 'application/x-bzip2':
            extensions = ['bz2', 'tbz2']
        else:
            cls = self.class_handler
            extensions = [ x[1:] for x in guess_all_extensions(format) ]
            if cls.class_extension in extensions:
                extensions.remove(cls.class_extension)
            extensions.insert(0, cls.class_extension)
        return extensions


    def get_handler(self):
        # Already loaded
        if self._handler is not None:
            return self._handler

        # Not yet loaded
        database = self.metadata.database
        base = self.metadata.uri
        cls = self.class_handler

        # Check the handler exists
        extensions = self.get_all_extensions()
        for extension in extensions:
            name = FileName.encode((self.name, extension, None))
            uri = base.resolve(name)
            # Found
            if database.has_handler(uri):
                self._handler = database.get_handler(uri, cls=cls)
                return self._handler

        # Not found, build a dummy one
        name = FileName.encode((self.name, cls.class_extension, None))
        uri = base.resolve(name)
        handler = cls()
        handler.database = database
        handler.uri = uri
        handler.timestamp = None
        handler.dirty = datetime.now()
        database.add_to_cache(uri, handler)
        self._handler = handler
        return self._handler

    handler = property(get_handler, None, None, '')


    def rename_handlers(self, new_name):
        folder = self.parent.handler
        old_name = self.name
        for extension in self.get_all_extensions():
            old = FileName.encode((old_name, extension, None))
            if folder.has_handler(old):
                return [(old, FileName.encode((new_name, extension, None)))]
        return None, None


    #######################################################################
    # Metadata
    #######################################################################
    @classmethod
    def get_metadata_schema(cls):
        schema = VersioningAware.get_metadata_schema()
        schema.update(WorkflowAware.get_metadata_schema())
        schema['filename'] = String
        return schema


    #######################################################################
    # Versioning & Indexing
    #######################################################################
    def to_text(self):
        return self.handler.to_text()


    def get_size(self):
        sizes = [ len(x.to_str()) for x in self.get_handlers() ]
        # XXX Maybe not the good algo
        return max(sizes)


    #######################################################################
    # User Interface
    #######################################################################
    def get_human_size(self):
        file = self.handler
        bytes = len(file.to_str())
        size = bytes / 1024.0
        if size >= 1024:
            size = size / 1024.0
            str = MSG(u'%.01f MB')
        else:
            str = MSG(u'%.01f KB')

        return str.gettext() % size


    def get_content_type(self):
        return self.handler.get_mimetype()

    # Views
    new_instance = File_NewInstance()
    download = File_Download()
    view = File_View()
    edit = File_Edit()
    externaledit = STLView(
        access='is_allowed_to_edit',
        title=MSG(u'External Editor'),
        icon='external.png',
        template='/ui/file/externaledit.xml')
    external_edit = File_ExternalEdit()
    backlinks = File_Backlinks()



###########################################################################
# Media
###########################################################################
class Image(File):
    class_id = 'image'
    class_version = '20071216'
    class_title = MSG(u'Image')
    class_icon16 = 'icons/16x16/image.png'
    class_icon48 = 'icons/48x48/image.png'
    class_views = ['view', 'download', 'edit', 'externaledit', 'upload',
                   'backlinks', 'edit_state', 'history']
    class_handler = ImageHandler

    # Views
    thumb = Image_Thumbnail()
    view = Image_View()



class Video(File):
    class_id = 'video'
    class_version = '20071216'
    class_title = MSG(u'Video')
    class_description = MSG(u'Video')
    class_icon16 = 'icons/16x16/flash.png'
    class_icon48 = 'icons/48x48/flash.png'

    # Views
    view = Video_View()



class Flash(File):
    class_id = 'application/x-shockwave-flash'
    class_version = '20071216'
    class_title = MSG(u'Flash')
    class_description = MSG(u'Flash Document')
    class_icon16 = 'icons/16x16/flash.png'
    class_icon48 = 'icons/48x48/flash.png'

    # Views
    view = Flash_View()



###########################################################################
# Office Documents
###########################################################################
class MSWord(File):
    class_id = 'application/msword'
    class_version = '20071216'
    class_title = MSG(u'Word')
    class_description = MSG(u'Word Text')
    class_icon16 = 'icons/16x16/word.png'
    class_icon48 = 'icons/48x48/word.png'
    class_handler = MSWordFile



class MSExcel(File):
    class_id = 'application/vnd.ms-excel'
    class_version = '20071216'
    class_title = MSG(u'Excel')
    class_description = MSG(u'Excel Spreadsheet')
    class_icon16 = 'icons/16x16/excel.png'
    class_icon48 = 'icons/48x48/excel.png'
    class_handler = MSExcelFile



class MSPowerPoint(File):
    class_id = 'application/vnd.ms-powerpoint'
    class_version = '20071216'
    class_title = MSG(u'PowerPoint')
    class_description = MSG(u'PowerPoint Presentation')
    class_icon16 = 'icons/16x16/powerpoint.png'
    class_icon48 = 'icons/48x48/powerpoint.png'
    class_handler = MSPowerPointFile



class OOWriter(File):
    class_id = 'application/vnd.sun.xml.writer'
    class_version = '20071216'
    class_title = MSG(u'OOo Writer')
    class_description = MSG(u'OpenOffice.org Text')
    class_icon16 = 'icons/16x16/oowriter.png'
    class_icon48 = 'icons/48x48/oowriter.png'
    class_handler = SXWFile



class OOCalc(File):
    class_id = 'application/vnd.sun.xml.calc'
    class_version = '20071216'
    class_title = MSG(u'OOo Calc')
    class_description = MSG(u'OpenOffice.org Spreadsheet')
    class_icon16 = 'icons/16x16/oocalc.png'
    class_icon48 = 'icons/48x48/oocalc.png'
    class_handler = SXCFile



class OOImpress(File):
    class_id = 'application/vnd.sun.xml.impress'
    class_version = '20071216'
    class_title = MSG(u'OOo Impress')
    class_description = MSG(u'OpenOffice.org Presentation')
    class_icon16 = 'icons/16x16/ooimpress.png'
    class_icon48 = 'icons/48x48/ooimpress.png'
    class_handler = SXIFile



class PDF(File):
    class_id = 'application/pdf'
    class_version = '20071216'
    class_title = MSG(u'PDF')
    class_description = MSG(u'PDF Document')
    class_icon16 = 'icons/16x16/pdf.png'
    class_icon48 = 'icons/48x48/pdf.png'
    class_handler = PDFFile



class RTF(File):
    class_id = 'text/rtf'
    class_version = '20071216'
    class_title = MSG(u"RTF")
    class_description = MSG(u'RTF Document')
    class_icon16 = 'icons/16x16/text.png'
    class_icon48 = 'icons/48x48/text.png'
    class_handler = RTFFile



class ODT(File):
    class_id = 'application/vnd.oasis.opendocument.text'
    class_version = '20071216'
    class_title = MSG(u'ODT')
    class_description = MSG(u'OpenDocument Text')
    class_icon16 = 'icons/16x16/odt.png'
    class_icon48 = 'icons/48x48/odt.png'
    class_handler = ODTFile



class ODS(File):
    class_id = 'application/vnd.oasis.opendocument.spreadsheet'
    class_version = '20071216'
    class_title = MSG(u'ODS')
    class_description = MSG(u'OpenDocument Spreadsheet')
    class_icon16 = 'icons/16x16/ods.png'
    class_icon48 = 'icons/48x48/ods.png'
    class_handler = ODSFile



class ODP(File):
    class_id = 'application/vnd.oasis.opendocument.presentation'
    class_version = '20071216'
    class_title = MSG(u'ODP')
    class_description = MSG(u'OpenDocument Presentation')
    class_icon16 = 'icons/16x16/odp.png'
    class_icon48 = 'icons/48x48/odp.png'
    class_handler = ODPFile



###########################################################################
# Archive Files
###########################################################################
class Archive(File):

    view = Archive_View()



class ZipArchive(Archive):
    class_id = 'application/zip'
    class_version = '20071216'
    class_title = MSG(u"Zip")
    class_description = MSG(u'Zip Archive')
    class_icon16 = 'icons/16x16/zip.png'
    class_icon48 = 'icons/48x48/zip.png'
    class_handler = ZIPFile



class TarArchive(Archive):
    class_id = 'application/x-tar'
    class_version = '20071216'
    class_title = MSG(u"Tar")
    class_description = MSG(u'Tar Archive')
    class_icon16 = 'icons/16x16/tar.png'
    class_icon48 = 'icons/48x48/tar.png'
    class_handler = TARFile



class Gzip(File):
    class_id = 'application/x-gzip'
    class_version = '20071216'
    class_title = MSG(u"Gzip")
    class_description = MSG(u'Gzip Compressed')
    class_icon16 = 'icons/16x16/gzip.png'
    class_icon48 = 'icons/48x48/gzip.png'
    class_handler = GzipFile



class Bzip2(File):
    class_id = 'application/x-bzip2'
    class_version = '20071216'
    class_title = MSG(u"Bzip2")
    class_description = MSG(u'Bzip2 Compressed')
    class_icon16 = 'icons/16x16/bzip.png'
    class_icon48 = 'icons/48x48/bzip.png'
    class_handler = Bzip2File


###########################################################################
# Register
###########################################################################
register_resource_class(File)
register_resource_class(File, format="application/octet-stream")
# Media
for cls in Image, Video, Flash:
    register_resource_class(cls)
# Office
for cls in MSWord, MSExcel, MSPowerPoint, PDF, RTF:
    register_resource_class(cls)
# OpenOffice 1.0 & ODF
for cls in OOWriter, OOCalc, OOImpress, ODT, ODS, ODP:
    register_resource_class(cls)
# Archives
for cls in ZipArchive, TarArchive, Gzip, Bzip2:
    register_resource_class(cls)




# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2008 Gautier Hayoun <gautier.hayoun@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from itools
from itools.gettext import MSG

# Import from ikaaro
from ikaaro.file import File
from ikaaro.folder import Folder
from ikaaro.folder_views import GoToSpecificDocument
from ikaaro.registry import register_resource_class
from ikaaro.resource_views import DBResource_Edit
from folder_views import WikiMenu
from page import WikiPage


class WikiFolder(Folder):

    class_id = 'WikiFolder'
    class_version = '20071215'
    class_title = MSG(u"Wiki")
    class_description = MSG(u"Container for a wiki")
    class_icon16 = 'wiki/WikiFolder16.png'
    class_icon48 = 'wiki/WikiFolder48.png'
    class_views = ['view', 'browse_content', 'preview_content',
                   'edit', 'new_resource', 'orphans', 'last_changes']

    __fixed_handlers__ = ['FrontPage']


    @staticmethod
    def _make_resource(cls, folder, name):
        Folder._make_resource(cls, folder, name)
        # FrontPage
        metadata = WikiPage.build_metadata(title={'en': u"Front Page"})
        folder.set_handler('%s/FrontPage.metadata' % name, metadata)


    def get_document_types(self):
        return [WikiPage, File]


    # User Interface
    context_menus = [WikiMenu()]

    view = GoToSpecificDocument(specific_document='FrontPage')
    edit = DBResource_Edit(title=MSG(u"Edit Wiki"))



###########################################################################
# Register
###########################################################################
register_resource_class(WikiFolder)



# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2008 Gautier Hayoun <gautier.hayoun@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from re import compile

# Import from docutils
from docutils import nodes
from docutils.core import publish_doctree
from docutils.readers import get_reader_class
from docutils.utils import SystemMessage

# Import from itools
from itools.gettext import MSG
from itools.handlers import checkid
from itools.uri import get_reference
from itools.web import get_context

# Import from ikaaro
from ikaaro.text import Text
from ikaaro.registry import register_resource_class
from ikaaro.resource_views import DBResource_NewInstance
from page_views import WikiPage_Edit, WikiPage_Help, WikiPage_ToPDF
from page_views import WikiPage_View



StandaloneReader = get_reader_class('standalone')


class WikiPage(Text):

    class_id = 'WikiPage'
    class_version = '20081114'
    class_title = MSG(u"Wiki Page")
    class_description = MSG(u"Wiki contents")
    class_icon16 = 'wiki/WikiPage16.png'
    class_icon48 = 'wiki/WikiPage48.png'
    class_views = ['view', 'to_pdf', 'edit', 'externaledit', 'upload',
                   'backlinks', 'edit_state', 'help']

    overrides = {
        # Security
        'file_insertion_enabled': 0,
        'raw_enabled': 0,
        # Encodings
        'input_encoding': 'utf-8',
        'output_encoding': 'utf-8',
    }


    #######################################################################
    # API
    #######################################################################
    def resolve_link(self, title):
        parent = self.parent

        # Try regular resource name or path
        try:
            return parent.get_resource(title)
        except (LookupError, UnicodeEncodeError):
            # Convert wiki name to resource name
            name = checkid(title)
            if name is None:
                return None
            try:
                return parent.get_resource(name)
            except LookupError:
                return None


    def get_document(self):
        parent = self.parent

        # Override dandling links handling
        class WikiReader(StandaloneReader):
            supported = ('wiki',)

            def wiki_reference_resolver(target):
                title = target['name']
                resource = self.resolve_link(title)
                if resource is None:
                    # Not Found
                    target['wiki_name'] = False
                else:
                    # Found
                    target['wiki_name'] = str(self.get_pathto(resource))

                return True

            wiki_reference_resolver.priority = 851
            unknown_reference_resolvers = [wiki_reference_resolver]

        # Publish!
        reader = WikiReader(parser_name='restructuredtext')
        document = publish_doctree(self.handler.to_str(), reader=reader,
                                   settings_overrides=self.overrides)

        context = get_context()
        # Assume internal paths are relative to the container
        for node in document.traverse(condition=nodes.reference):
            refuri = node.get('refuri')
            # Skip wiki or fragment link
            if node.get('wiki_name') or not refuri:
                continue
            reference = get_reference(refuri.encode('utf_8'))
            # Skip external
            if reference.scheme or reference.authority:
                continue
            # Note: absolute paths will be rewritten as relative paths
            try:
                resource = parent.get_resource(reference.path)
                node['refuri'] = context.get_link(resource)
            except LookupError:
                pass

        # Assume image paths are relative to the container
        for node in document.traverse(condition=nodes.image):
            uri  = node['uri'].encode('utf_8')
            reference = get_reference(uri)
            # Skip external
            if reference.scheme or reference.authority:
                continue
            # Strip the view
            path = reference.path
            if reference.path[-1] == ';download':
                path = path[:-1]
            # Get the resource
            try:
                resource = parent.get_resource(path)
            except LookupError:
                pass
            else:
                node['uri'] = '%s/;download' % context.get_link(resource)

        return document


    def get_links(self):
        base = self.get_abspath()

        links = []
        try:
            document = self.get_document()
        except SystemMessage:
            # The document is in a incoherent state
            return None
        for node in document.traverse(condition=nodes.reference):
            refname = node.get('wiki_name')
            if refname is False:
                # Wiki link not found
                title = node['name']
                path = checkid(title) or title
                path = base.resolve(path)
            elif refname:
                # Wiki link found, "refname" is the path
                path = base.resolve2(refname)
            else:
                # Regular link, include internal ones
                refuri = node.get('refuri')
                if refuri is None:
                    continue
                reference = get_reference(refuri.encode('utf_8'))
                # Skip external
                if reference.scheme or reference.authority:
                    continue
                path = base.resolve2(reference.path)
            path = str(path)
            links.append(path)

        for node in document.traverse(condition=nodes.image):
            uri = node['uri'].encode('utf_8')
            reference = get_reference(uri)
            # Skip external image
            if reference.scheme or reference.authority:
                continue
            # Strip the view
            path = reference.path
            if path[-1] == ';download':
                path = path[:-1]
            # Resolve the path
            path = base.resolve(path)
            path = str(path)
            links.append(path)

        return links


    def change_link(self, old_path, new_path,
                    links_re = compile(r'(\.\. .*?: )(\S*)')):
        old_data = self.handler.to_str()
        new_data = []

        not_uri = 0
        base = self.parent.get_abspath()
        for segment in links_re.split(old_data):
            not_uri = (not_uri + 1) % 3
            if not not_uri:
                reference = get_reference(segment)

                # Skip external link
                if reference.scheme or reference.authority:
                    new_data.append(segment)
                    continue

                # Strip the view
                path = reference.path
                if path[-1] == ';download':
                    path = path[:-1]
                    view = '/;download'
                else:
                    view = ''

                # Resolve the path
                path = base.resolve(path)

                # Match ?
                if path == old_path:
                    segment = str(base.get_pathto(new_path)) + view
            new_data.append(segment)
        new_data = ''.join(new_data)
        self.handler.load_state_from_string(new_data)
        get_context().server.change_resource(self)


    #######################################################################
    # Update service
    #######################################################################
    def update_20081114(self,
            links_migration_re = compile(r'\.\. figure:: ([^;]*?)(?!;)(\s)'),
            links_migration_sub = r'.. figure:: \1/;download\2'):
        data = self.handler.to_str()
        data = links_migration_re.sub(links_migration_sub, data)
        self.handler.load_state_from_string(data)


    #######################################################################
    # User Interface
    #######################################################################
    new_instance = DBResource_NewInstance()
    view = WikiPage_View()
    to_pdf = WikiPage_ToPDF()
    edit = WikiPage_Edit()
    help = WikiPage_Help()


    def get_context_menus(self):
        return self.parent.get_context_menus()



###########################################################################
# Register
###########################################################################
register_resource_class(WikiPage)



# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007 Luis Arturo Belmar-Letelier <luis@itaapy.com>
# Copyright (C) 2007-2008 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007-2008 Nicolas Deram <nicolas@itaapy.com>
# Copyright (C) 2007-2008 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2008 Gautier Hayoun <gautier.hayoun@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from datetime import datetime
from string import Template

# Import from itools
from itools.csv import Table
from itools.datatypes import DateTime, Integer, String, Unicode, Tokens
from itools.gettext import MSG
from itools.handlers import checkid
from itools.vfs import FileName
from itools.xapian import IntegerField, KeywordField
from itools.web import get_context

# Import from ikaaro
from ikaaro.file import File
from ikaaro.folder import Folder
from ikaaro.registry import register_resource_class, get_resource_class
from ikaaro.utils import generate_name
from issue_views import Issue_Edit, Issue_EditResources, Issue_History
from issue_views import IssueTrackerMenu


class History(Table):

    record_schema = {
        'datetime': DateTime,
        'username': String,
        'title': Unicode,
        'product': Integer,
        'module': Integer,
        'version': Integer,
        'type': Integer,
        'state': Integer,
        'priority': Integer,
        'assigned_to': String,
        'comment': Unicode,
        'cc_list': Tokens(),
        'file': String}



class Issue(Folder):

    class_id = 'issue'
    class_version = '20071216'
    class_title = MSG(u'Issue')
    class_description = MSG(u'Issue')
    class_views = ['edit', 'edit_resources', 'browse_content', 'history']


    @staticmethod
    def _make_resource(cls, folder, name):
        Folder._make_resource(cls, folder, name)
        folder.set_handler('%s/.history' % name, History())


    def get_catalog_fields(self):
        fields = Folder.get_catalog_fields(self)
        # Metadata
        names = [
            'id', 'product', 'module', 'version', 'type', 'state', 'priority']
        for name in names:
            field = IntegerField(name, is_stored=True)
            fields.append(field)
        # Assign To
        field = KeywordField('assigned_to', is_stored=True)
        fields.append(field)
        # Ok
        return fields


    def get_catalog_values(self):
        document = Folder.get_catalog_values(self)
        document['id'] = int(self.name)
        names = 'product', 'module', 'version', 'type', 'priority', 'state'
        for name in names:
            document[name] = self.get_value(name)
        document['assigned_to'] = self.get_value('assigned_to') or 'nobody'
        document['title'] = self.get_value('title')
        return document


    def get_document_types(self):
        return [File]


    def get_mtime(self):
        """Return the datetime of the last record.
        """
        history = self.get_history()
        record = history.get_record(-1)
        if record:
            return history.get_record_value(record, 'datetime')
        return self.get_mtime()


    def get_links(self):
        base = self.get_abspath()

        links = []
        for record in self.get_history_records():
            filename = record.file
            if filename:
                links.append(str(base.resolve2(filename)))
        return links


    def change_link(self, old_path, new_path):
        base = self.get_abspath()
        old_name = base.get_pathto(old_path)
        history = self.get_history()
        for record in history.get_records():
            file = history.get_record_value(record, 'file')
            if file == old_name:
                value = str(base.get_pathto(new_path))
                history.update_record(record.id, **{'file': value})
        # Reindex
        get_context().server.change_resource(self)


    #######################################################################
    # API
    #######################################################################
    def get_title(self, language=None):
        return '#%s %s' % (self.name, self.get_value('title'))


    def get_calendar(self):
        return self.parent.get_resource('calendar')


    def load_handlers(self):
        Folder.load_handlers(self)
        self.get_history()


    def get_history(self):
        return self.handler.get_handler('.history', cls=History)


    def get_history_records(self):
        return self.get_history().get_records()


    def get_value(self, name):
        history = self.get_history()
        record = history.get_record(-1)
        if record:
            return history.get_record_value(record, name)
        return None


    def _add_record(self, context, form):
        user = context.user
        root = context.root
        parent = self.parent
        users = root.get_resource('users')

        record = {}
        # Datetime
        record['datetime'] = datetime.now()
        # User XXX
        if user is None:
            record['username'] = ''
        else:
            record['username'] = user.name
        # Title
        title = context.get_form_value('title', type=Unicode).strip()
        record['title'] = title
        # Version, Priority, etc.
        for name in ['product', 'module', 'version', 'type', 'state',
                     'priority', 'assigned_to', 'comment']:
            type = History.record_schema[name]
            value = context.get_form_value(name, type=type)
            if type == Unicode:
                value = value.strip()
            record[name] = value
        # CCs
        cc_list = set(self.get_value('cc_list') or ())
        cc_remove = context.get_form_value('cc_remove')
        if cc_remove:
            cc_remove = context.get_form_values('cc_list')
            cc_list = cc_list.difference(cc_remove)
        cc_add = context.get_form_values('cc_add')
        if cc_add:
            cc_list = cc_list.union(cc_add)
        record['cc_list'] = list(cc_list)

        # Files XXX
        file = context.get_form_value('file')
        if file is None:
            record['file'] = ''
        else:
            # Upload
            filename, mimetype, body = form['file']
            # Find a non used name
            name = checkid(filename)
            name, extension, language = FileName.decode(name)
            name = generate_name(name, self.get_names())
            # Add attachement
            cls = get_resource_class(mimetype)
            cls.make_resource(cls, self, name, body=body, filename=filename,
                            extension=extension, format=mimetype)
            # Link
            record['file'] = name
        # Update
        modifications = self.get_diff_with(record, context)
        history = self.get_history()
        history.add_record(record)
        # Send a Notification Email
        # Notify / From
        if user is None:
            user_title = MSG(u'ANONYMOUS')
        else:
            user_title = user.get_title()
        # Notify / To
        to_addrs = set()
        reported_by = self.get_reported_by()
        if reported_by:
            to_addrs.add(reported_by)
        for cc in cc_list:
            to_addrs.add(cc)
        assigned_to = self.get_value('assigned_to')
        if assigned_to:
            to_addrs.add(assigned_to)
        if user.name in to_addrs:
            to_addrs.remove(user.name)
        # Notify / Subject
        tracker_title = self.parent.get_property('title') or 'Tracker Issue'
        subject = '[%s #%s] %s' % (tracker_title, self.name, title)
        # Notify / Body
        if context.resource.class_id == 'tracker':
            uri = context.uri.resolve('%s/;edit' % self.name)
        else:
            uri = context.uri.resolve(';edit')
        body = '#%s %s %s\n\n' % (self.name, self.get_value('title'),
                                  str(uri))
        message = MSG(u'The user $title did some changes.')
        body +=  message.gettext(title=user_title)
        body += '\n\n'
        if file:
            filename = unicode(filename, 'utf-8')
            message = MSG(u'New Attachment: $filename')
            message = message.gettext(filename=filename)
            body += message + '\n'
        comment = context.get_form_value('comment', type=Unicode)
        if modifications:
            body += modifications
            body += '\n\n'
        if comment:
            title = MSG(u'Comment').gettext()
            separator = len(title) * u'-'
            template = u'${title}\n${separator}\n\n${comment}\n'
            template = Template(template)
            body += template.substitute(title=title, separator=separator,
                                        comment=comment)
        # Notify / Send
        for to_addr in to_addrs:
            user = root.get_user(to_addr)
            if not user:
                continue
            to_addr = user.get_property('email')
            root.send_email(to_addr, subject, text=body)


    def get_diff_with(self, record, context):
        """Return a text with the diff between the last and new issue state
        """
        root = context.root
        modifications = []
        history = self.get_history()
        if history.get_n_records() > 0:
            # Edit issue
            template = MSG(u'$field: $old_value to $new_value')
            empty = MSG(u'[empty]').gettext()
        else:
            # New issue
            template = MSG(u'$field: $old_value$new_value')
            empty = u''
        # Modification of title
        last_title = self.get_value('title') or empty
        new_title = record['title']
        if last_title != new_title:
            field = MSG(u'Title').gettext()
            text = template.gettext(field=field, old_value=last_title,
                                    new_value=new_title)
            modifications.append(text)
        # List modifications
        fields = [
            ('module', MSG(u'Module')),
            ('version', MSG(u'Version')),
            ('type', MSG(u'Type')),
            ('priority', MSG(u'Priority')),
            ('state', MSG(u'State'))]
        for name, field in fields:
            field = field.gettext()
            new_value = record[name]
            last_value = self.get_value(name)
            # Detect if modifications
            if last_value == new_value:
                continue
            new_title = last_title = empty
            csv = self.parent.get_resource(name).handler
            if last_value or last_value == 0:
                rec = csv.get_record(last_value)
                last_title = csv.get_record_value(rec, 'title')
            if new_value or new_value == 0:
                rec = csv.get_record(new_value)
                new_title = csv.get_record_value(rec, 'title')
            text = template.gettext(field=field, old_value=last_title,
                                    new_value=new_title)
            modifications.append(text)

        # Modifications of assigned_to
        last_user = self.get_value('assigned_to') or ''
        new_user = record['assigned_to']
        if last_user != new_user:
            if last_user:
                last_user = root.get_user(last_user).get_property('email')
            if new_user:
                new_user = root.get_user(new_user).get_property('email')
            field = MSG(u'Assigned To').gettext()
            text = template.gettext(field=field, old_value=last_user or empty,
                                    new_value=new_user or empty)
            modifications.append(text)

        # Modifications of cc_list
        last_cc = list(self.get_value('cc_list') or ())
        new_cc = list(record['cc_list'] or ())
        if last_cc != new_cc:
            last_values = []
            for cc in last_cc:
                value = root.get_user(cc).get_property('email')
                last_values.append(value)
            new_values = []
            for cc in new_cc:
                value = root.get_user(cc).get_property('email')
                new_values.append(value)
            field = MSG(u'CC').gettext()
            last_values = ', '.join(last_values) or empty
            new_values = ', '.join(new_values) or empty
            text = template.gettext(field=field, old_value=last_values,
                                    new_value=new_values)
            modifications.append(text)

        return u'\n'.join(modifications)


    def get_reported_by(self):
        history = self.get_history()
        return history.get_record(0).username


    def get_comment(self):
        records = list(self.get_history_records())
        i = len(records) - 1
        while i >= 0:
            record = records[i]
            comment = record.comment
            if comment:
                return comment
            i -= 1
        return ''


    def to_text(self):
        records = list(self.get_history_records())
        comments = [ r.comment for r in records
                     if r.comment ]
        return u'\n'.join(comments)


    def has_text(self, text):
        if text in self.get_value('title').lower():
            return True
        return text in self.get_comment().lower()


    def get_size(self):
        # FIXME Used by the browse list view (size is indexed)
        return 0


    #######################################################################
    # User Interface
    #######################################################################
    def get_context_menus(self):
        return self.parent.get_context_menus() + [IssueTrackerMenu()]


    edit = Issue_Edit()
    edit_resources = Issue_EditResources()
    history = Issue_History()




###########################################################################
# Register
###########################################################################
register_resource_class(Issue)



# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007 Nicolas Deram <nicolas@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from itools
from itools.gettext import MSG
from itools.web import INFO, ERROR



MSG_BAD_KEY = ERROR(u"Your confirmation key is invalid.")

MSG_BAD_NAME = ERROR(
    u'The document name contains illegal characters, choose another one.')

MSG_CAPTION = ERROR(u'Caption')

MSG_CHANGES_SAVED = INFO(u'The changes have been saved.')

MSG_CHANGES_SAVED2 = INFO(u'The changes have been saved ($time).')

MSG_DELETE_RESOURCE = MSG(u'Are you sure you want to delete this resource?')

MSG_DELETE_SELECTION = MSG(u'Are you sure you want to delete the selection?')

MSG_EDIT_CONFLICT = ERROR(
    u'Someone already saved this document, click "Save" again to force.')

MSG_EDIT_CONFLICT2 = ERROR(
    u'User $user already saved this document, click "Save" again to force.')

MSG_EMPTY_FILENAME = ERROR(u'The file must be entered.')

MSG_EXISTANT_FILENAME = ERROR(u'A given name already exists.')

MSG_INVALID_EMAIL = ERROR(u'The email address provided is invalid.')

MSG_NAME_CLASH = ERROR(u'There is already another resource with this name.')

MSG_NAME_MISSING = ERROR(u'The name is missing.')

MSG_NEW_RESOURCE = INFO(u'A new resource has been added.')

MSG_NONE_REMOVED = ERROR(u'No resource removed.')

MSG_RESOURCES_PASTED = INFO(u'Resources pasted: $resources.')

MSG_RESOURCES_REMOVED = INFO(u'Resources removed: $resources.')

MSG_RESOURCES_REFERENCED = ERROR(u'Action impossible '
                                 u'(the resources are in use): $resources.')

MSG_RESOURCES_NOT_PASTED = ERROR(u'Resources not allowed to paste here: '
                                 u'$resources.')

MSG_RESOURCES_NOT_REMOVED = ERROR(u'Resources not allowed to remove: '
                                  u'$resources.')

MSG_PAGE_LOCK = ERROR(u'This page is locked by $user')

MSG_PASSWORD_MISMATCH = ERROR(u'The provided passwords do not match.')

MSG_REGISTERED = ERROR(
    u"You have already confirmed your registration. "
    u"Try to log in or ask for a new password.")

MSG_NONE_SELECTED = ERROR(u'No resource selected.')

MSG_NONE_ALLOWED = ERROR(u"No resource allowed.")

MSG_NO_PASTE = ERROR(u'Nothing to paste.')

MSG_RENAMED = INFO(u'Resources renamed.')

MSG_COPIED = INFO(u'Resources copied.')

MSG_CUT = INFO(u'Resources cut.')

MSG_PUBLISHED = INFO(u'Resources published.')



# -*- coding: UTF-8 -*-
# Copyright (C) 2006-2007 Juan David Ibáñez Palomar <jdavid@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from datetime import datetime
from random import random
from time import time

# Import from itools
from itools.datatypes import DateTime
from itools.handlers import TextFile, register_handler_class
from itools.utils import add_type


class Lock(TextFile):

    class_mimetypes = ['text/x-lock']
    class_extension = 'lock'


    def new(self, username=None, **kw):
        self.username = username
        self.lock_timestamp = datetime.now()
        self.key = '%s-%s-00105A989226:%.03f' % (random(), random(), time())


    def _load_state_from_file(self, file):
        username, timestamp, key = file.read().strip().split('\n')
        self.username = username
        # FIXME Backwards compatibility: remove microseconds first
        timestamp = timestamp.split('.')[0]
        self.lock_timestamp = DateTime.decode(timestamp)
        self.key = key


    def to_str(self):
        timestamp = DateTime.encode(self.lock_timestamp)
        return '%s\n%s\n%s' % (self.username, timestamp, self.key)



###########################################################################
# Register
###########################################################################
register_handler_class(Lock)
for mimetype in Lock.class_mimetypes:
    add_type(mimetype, '.%s' % Lock.class_extension)



# -*- coding: UTF-8 -*-
# Copyright (C) 2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from base64 import decodestring, encodestring
from marshal import dumps, loads
from urllib import quote, unquote
from zlib import compress, decompress

# Import from itools
from itools.datatypes import DataType, Enumerate
from itools.utils import freeze, guess_type


"""This module defines some datatypes used in ikaaro, whose inclusion in
itools is not yet clear.
"""


class FileDataType(DataType):
    """FIXME This datatype is special in that it does not deserializes from
    a byte string, but from a tuple.  Some day we should find a correct
    solution.
    """
    @staticmethod
    def encode(value):
        """Cannot preload anything in a file input.
        """
        return None


    @staticmethod
    def decode(data):
        """Find out the resource class (the mimetype sent by the browser can be
        minimalistic).
        """
        filename, mimetype, body = data
        # Find out the mimetype
        guessed, encoding = guess_type(filename)
        if encoding is not None:
            encoding_map = {'gzip': 'application/x-gzip',
                            'bzip2': 'application/x-bzip2'}
            if encoding in encoding_map:
                mimetype = encoding_map[encoding]
        elif guessed is not None:
            mimetype = guessed

        return filename, mimetype, body



class Password(DataType):

    @staticmethod
    def decode(data):
        return decodestring(unquote(data))


    @staticmethod
    def encode(value):
        return quote(encodestring(value))



class CopyCookie(DataType):

    default = None, freeze([])

    @staticmethod
    def encode(value):
        return quote(compress(dumps(value), 9))


    @staticmethod
    def decode(str):
        return loads(decompress(unquote(str)))



class ImageWidth(Enumerate):
    options = [{'name': '640', 'value': u"small"},
               {'name': '800', 'value': u"medium"},
               {'name': '1024', 'value': u"large"},
               {'name': '1280', 'value': u"huge"},
               {'name': '', 'value': u"original"}]



# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007-2008 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2008 Nicolas Deram <nicolas@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from itools
import itools
from itools import get_abspath
from itools.datatypes import Email, String, Unicode
from itools.datatypes import Enumerate
from itools.gettext import MSG
from itools.handlers import merge_dicts
from itools.stl import stl
from itools import vfs
from itools.web import STLView, INFO, ERROR
from itools.xapian import PhraseQuery, OrQuery, AndQuery, TextField

# Import from ikaaro
import ikaaro
from forms import AutoForm, SelectWidget, MultilineWidget, TextWidget
import messages
from registry import get_resource_class
from registry import get_register_websites, get_website_class
from resource_views import AddResourceMenu
from views import NewInstanceForm, SearchForm
from utils import get_base_path_query



class NewWebSiteForm(NewInstanceForm):

    access = 'is_allowed_to_add'
    title = MSG(u'Web Site')
    template = '/ui/website/new_instance.xml'
    schema = {
        'name': String,
        'title': Unicode,
        'class_id': String}
    context_menus = [AddResourceMenu()]


    def get_namespace(self, resource, context):
        type = context.get_query_value('type')
        cls = get_resource_class(type)

        # Specific Websites
        websites = get_register_websites()
        websites = list(websites)
        if len(websites) == 1:
            websites = None
        else:
            selected = context.get_form_value('class_id')
            websites = [
                {'title': x.class_title.gettext(),
                 'class_id': x.class_id,
                 'selected': x.class_id == selected,
                 'icon': '/ui/' + x.class_icon16}
                for x in websites ]
            if selected is None:
                websites[0]['selected'] = True

        # Ok
        return {
            'class_id': cls.class_id,
            'class_title': cls.class_title.gettext(),
            'websites': websites}


    def get_new_resource_name(self, form):
        # If the name is not explicitly given, use the title
        return form['name'].strip() or form['title'].strip()


    def action(self, resource, context, form):
        name = form['name']
        title = form['title']
        class_id = form['class_id']

        # Find out the class id
        if class_id is None:
            websites = get_register_websites()
            websites = list(websites)
            class_id = websites[0].class_id

        # Make resource
        cls = get_website_class(class_id)
        child = cls.make_resource(cls, resource, name)

        # Add title
        metadata = child.metadata
        language = resource.get_site_root().get_default_language()
        metadata.set_property('title', title, language=language)

        goto = './%s/' % name
        return context.come_back(messages.MSG_NEW_RESOURCE, goto=goto)



class ForgottenPasswordForm(AutoForm):

    access = True
    title = MSG(u'Forgotten password')
    submit_value = MSG(u'Ok')

    widgets = [
        TextWidget('username', title=MSG(u'Type your email address')),
        ]

    schema = query_schema = {'username': Email(default='')}


    def get_value(self, resource, context, name, datatype):
        if name == 'username':
            return context.get_query_value('username')
        return AutoForm.get_value(self, resource, context, name, datatype)


    def action(self, resource, context, form):
        username = form['username']
        # TODO Don't generate the password, send instead a link to a form
        # where the user will be able to type his new password.
        root = context.root

        # Get the email address
        username = username.strip()

        # Get the user with the given login name
        results = root.search(username=username)
        if results.get_n_documents() == 0:
            message = ERROR(u'There is not a user identified as "$username"',
                      username=username)
            context.message = message
            return

        user = results.get_documents()[0]
        user = resource.get_resource('/users/%s' % user.name)

        # Send email of confirmation
        email = user.get_property('email')
        user.send_forgotten_password(context, email)

        handler = resource.get_resource('/ui/website/forgotten_password.xml')
        return stl(handler)



class RegisterForm(AutoForm):

    access = 'is_allowed_to_register'
    title = MSG(u'Register')
    submit_value = MSG(u'Register')

    schema = {
        'firstname': Unicode(mandatory=True),
        'lastname': Unicode(mandatory=True),
        'email': Email(mandatory=True)}

    widgets = [
        TextWidget('firstname', title=MSG(u'First Name')),
        TextWidget('lastname', title=MSG(u'Last Name')),
        TextWidget('email', title=MSG(u'E-mail Address'))]


    def action(self, resource, context, form):
        # Get input data
        firstname = form['firstname'].strip()
        lastname = form['lastname'].strip()
        email = form['email'].strip()

        # Do we already have a user with that email?
        root = context.root
        user = root.get_user_from_login(email)
        if user is not None:
            if not user.has_property('user_must_confirm'):
                message = u'There is already an active user with that email.'
                context.message = ERROR(message)
                return
        else:
            # Add the user
            users = resource.get_resource('users')
            user = users.set_user(email, None)
            user.set_property('firstname', firstname)
            user.set_property('lastname', lastname)
            # Set the role
            default_role = resource.__roles__[0]['name']
            resource.set_user_role(user.name, default_role)

        # Send confirmation email
        user.send_confirmation(context, email)

        # Bring the user to the login form
        message = MSG(
            u"An email has been sent to you, to finish the registration "
            u"process follow the instructions detailed in it.")
        return message.gettext().encode('utf-8')



class ContactOptions(Enumerate):

    @classmethod
    def get_options(cls):
        resource = cls.resource
        users = resource.get_resource('/users')

        return [
            {'name': x, 'value': users.get_resource(x).get_title()}
            for x in resource.get_property('contacts') ]



class ContactForm(AutoForm):

    access = True
    title = MSG(u'Contact')
    submit_value = MSG(u'Send')
    query_schema = {'to': String,
                    'subject': Unicode,
                    'body': Unicode}

    def get_schema(self, resource, context):
        return {
            'to': ContactOptions(resource=resource, mandatory=True),
            'from': Email(mandatory=True),
            'subject': Unicode(mandatory=True),
            'body': Unicode(mandatory=True),
        }


    widgets = [
        SelectWidget('to', title=MSG(u'Recipient')),
        TextWidget('from', title=MSG(u'Your email address'), size=40),
        TextWidget('subject', title=MSG(u'Message subject'), size=40),
        MultilineWidget('body', title=MSG(u'Message body'), rows=8, cols=50),
    ]


    def get_value(self, resource, context, name, datatype):
        if name == 'from':
            user = context.user
            if user is not None:
                return user.get_property('email')
        else:
            query = context.query
            if name in query:
                return query[name]
        return datatype.get_default()


    def action(self, resource, context, form):
        # Get form values
        contact = form['to']
        from_addr = form['from'].strip()
        subject = form['subject'].strip()
        body = form['body'].strip()

        # Find out the "to" address
        contact = resource.get_resource('/users/%s' % contact)
        contact_title = contact.get_title()
        contact = contact.get_property('email')
        if contact_title != contact:
            contact = (contact_title, contact)
        # Send the email
        root = resource.get_root()
        root.send_email(contact, subject, from_addr=from_addr, text=body)
        # Ok
        context.message = INFO(u'Message sent.')



class SiteSearchView(SearchForm):

    access = True
    title = MSG(u'Search')
    template = '/ui/website/search.xml'

    search_schema = {
        'site_search_text': Unicode,
    }


    def get_namespace(self, resource, context):
        namespace = SearchForm.get_namespace(self, resource, context)
        namespace['text'] = context.query['site_search_text'].strip()
        return namespace


    def get_query_schema(self):
        schema = merge_dicts(SearchForm.get_query_schema(self))
        del schema['sort_by']
        del schema['reverse']
        return schema


    def get_search_namespace(self, resource, context):
        text = context.query['site_search_text']
        return {'text': text}


    search_template = '/ui/website/search_form.xml'
    def get_items(self, resource, context):
        text = context.query['site_search_text'].strip()
        if not text:
            return []

        # The Search Query
        query = [ OrQuery(PhraseQuery('title', word),
                  PhraseQuery('text', word))
                    for word, kk in TextField.split(text) ]
        if not query:
            return []

        # Search
        abspath = resource.get_canonical_path()
        q1= get_base_path_query(str(abspath))
        query = AndQuery(q1, *query)
        root = context.root
        results = root.search(query=query)
        documents = results.get_documents()

        # Check access rights
        user = context.user
        items = []
        for document in documents:
            child = root.get_resource(document.abspath)
            ac = child.get_access_control()
            if ac.is_allowed_to_view(user, child):
                items.append(child)

        return items


    def sort_and_batch(self, resource, context, items):
        # Batch
        start = context.query['batch_start']
        size = context.query['batch_size']
        return items[start:start+size]


    table_template = '/ui/website/search_table.xml'
    def get_table_namespace(self, resource, context, items):
        # Build the namespace
        site_root = resource.get_site_root()
        items_ns = [{
            'abspath': '/%s' % site_root.get_pathto(item),
            'title': item.get_title(),
            'type': item.class_title.gettext(),
            'size': item.get_human_size(),
            'url': '%s/' % resource.get_pathto(item),
            'icon': item.get_class_icon(),
        } for item in items ]

        return {'items': items_ns}



class AboutView(STLView):

    access = True
    title = MSG(u'About')
    template = '/ui/root/about.xml'


    def get_namespace(self, resource, context):
        return  {
            'itools_version': itools.__version__,
            'ikaaro_version': ikaaro.__version__,
        }



class CreditsView(STLView):

    access = True
    title = MSG(u'Credits')
    template = '/ui/root/credits.xml'


    def get_namespace(self, resource, context):
        context.styles.append('/ui/credits.css')

        # Build the namespace
        credits = get_abspath('CREDITS')
        lines = vfs.open(credits).readlines()
        names = [ x[3:].strip() for x in lines if x.startswith('N: ') ]

        return {'hackers': names}




# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2006-2007 Nicolas Deram <nicolas@itaapy.com>
# Copyright (C) 2006-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from itools
from itools.gettext import MSG
from itools.handlers import Folder as FolderHandler
from itools import vfs
from itools.web import get_context, BaseView
from itools.xapian import PhraseQuery

# Import from ikaaro
from exceptions import ConsistencyError
from folder_views import Folder_BrowseContent, Folder_LastChanges
from folder_views import Folder_NewResource, Folder_Orphans, Folder_Thumbnail
from folder_views import Folder_PreviewContent, Folder_Rename, Folder_View
from registry import register_resource_class, get_resource_class
from resource_ import DBResource



class Folder(DBResource):

    class_id = 'folder'
    class_version = '20071215'
    class_title = MSG(u'Folder')
    class_description = MSG(u'Organize your files and documents with folders.')
    class_icon16 = 'icons/16x16/folder.png'
    class_icon48 = 'icons/48x48/folder.png'
    class_views = ['view', 'browse_content', 'preview_content', 'edit']
    class_handler = FolderHandler


    #########################################################################
    # Aggregation relationship (what a generic folder can contain)
    class_document_types = []

    __fixed_handlers__ = []


    #########################################################################
    # Gallery properties
    DEFAULT_SIZE = 128
    MIN_SIZE = 32
    MAX_SIZE = 512
    SIZE_STEPS = (32, 48, 64, 128, 256, 512)


    @classmethod
    def register_document_type(cls, handler_class):
        cls.class_document_types.append(handler_class)


    def get_document_types(self):
        return self.class_document_types


    #######################################################################
    # API
    #######################################################################
    def _get_names(self):
        folder = self.handler
        return [ x[:-9] for x in folder.get_handler_names()
                 if x[-9:] == '.metadata' ]


    def _get_resource(self, name):
        folder = self.handler
        metadata = folder.get_handler('%s.metadata' % name)
        format = metadata.format

        uri = folder.uri.resolve2(name)
        if vfs.exists(uri):
            is_file = vfs.is_file(uri)
        else:
            # FIXME This is just a guess, it may fail.
            is_file = '/' in format

        cls = get_resource_class(format, is_file=is_file)
        return cls(metadata)


    def update_links(self, new_name, base_path):
        """The resource must update its links to itself and to its content
        """
        # Check referencial-integrity
        catalog = get_context().server.catalog
        # The catalog is not available when updating (icms-update.py)
        # FIXME We do not guarantee referencial-integrity when updating
        if catalog is None:
            return

        old_base_path = self.get_abspath()
        new_base_path = base_path.resolve2(new_name)
        for resource in self.traverse_resources():
            # Old and new paths
            old_path = resource.get_abspath()
            pathto = old_base_path.get_pathto(old_path)
            new_path = new_base_path.resolve2(pathto)

            # Get all the resources that have a link to this resource
            query = PhraseQuery('links', str(old_path))
            results = catalog.search(query).get_documents()
            for result in results:
                resource = self.get_resource(result.abspath)
                resource.change_link(old_path, new_path)


    def del_resource(self, name):
        context = get_context()
        resource = self.get_resource(name)

        # Check referencial-integrity
        catalog = context.server.catalog
        # The catalog is not available when updating (icms-update.py)
        # FIXME We do not guarantee referencial-integrity when updating
        if catalog is not None:
            # FIXME Check sub-resources too
            path = str(resource.get_abspath())
            results = catalog.search(links=path)
            if results.get_n_documents() > 0:
                message = 'cannot delete, resource "%s" is referenced' % path
                raise ConsistencyError, message

        # Events, remove
        context.server.remove_resource(resource)
        # Remove
        folder = self.handler
        folder.del_handler('%s.metadata' % name)
        for handler in resource.get_handlers():
            if folder.has_handler(handler.uri):
                folder.del_handler(handler.uri)


    def _resolve_source_target(self, source, target):
        # Find out the source and target absolute URIs
        folder = self.handler
        if source[0] == '/':
            source_uri = self.get_root().handler.uri.resolve2(source[1:])
        else:
            source_uri = folder.uri.resolve2(source)
        if target[0] == '/':
            target_uri = self.get_root().handler.uri.resolve2(target[1:])
        else:
            target_uri = folder.uri.resolve2(target)

        # Load the handlers so they are of the right class, for resources
        # like that define explicitly the handler class.  This fixes for
        # instance copy&cut&paste of a tracker in a just started server.
        # TODO this is a work-around, there should be another way to define
        # explicitly the handler class.
        resource = self.get_resource(source)
        if isinstance(resource, Folder):
            for resource in resource.traverse_resources():
                resource.load_handlers()
        else:
            resource.load_handlers()

        return source_uri, target_uri


    def copy_resource(self, source, target):
        server = get_context().server

        # Find out the source and target absolute URIs
        source_uri, target_uri = self._resolve_source_target(source, target)
        new_name = target_uri.path[-1]

        # Get the source resource
        resource = self.get_resource(source)

        # Copy the metadata
        folder = self.handler
        folder.copy_handler('%s.metadata' % source_uri,
                            '%s.metadata' % target_uri)
        # Copy the content
        for old_name, new_name in resource.rename_handlers(new_name):
            if old_name is None:
                continue
            src_uri = source_uri.resolve(old_name)
            dst_uri = target_uri.resolve(new_name)
            if folder.has_handler(src_uri):
                folder.copy_handler(src_uri, dst_uri)

        # Events, add
        resource = self.get_resource(target)
        server.add_resource(resource)


    def move_resource(self, source, target):
        server = get_context().server

        # Find out the source and target absolute URIs
        source_uri, target_uri = self._resolve_source_target(source, target)
        new_name = target_uri.path[-1]

        # Get the source resource
        resource = self.get_resource(source)

        # Update the links to the resources that are to be moved
        resource.update_links(target, self.get_abspath())
        # Events, remove
        server.remove_resource(resource)

        # Move the metadata
        folder = self.handler
        folder.move_handler('%s.metadata' % source_uri,
                            '%s.metadata' % target_uri)
        # Move the content
        for old_name, new_name in resource.rename_handlers(new_name):
            if old_name is None:
                continue
            src_uri = source_uri.resolve(old_name)
            dst_uri = target_uri.resolve(new_name)
            if folder.has_handler(src_uri):
                folder.move_handler(src_uri, dst_uri)

        # Events, add
        resource = self.get_resource(target)
        server.add_resource(resource)


    def traverse_resources(self):
        yield self
        for name in self._get_names():
            resource = self.get_resource(name)
            if isinstance(resource, Folder):
                for x in resource.traverse_resources():
                    yield x
            else:
                yield resource


    def search_resources(self, cls=None, format=None, state=None):
        if cls is None:
            cls = DBResource

        for resource in self.get_resources():
            # Filter by base class
            if not isinstance(resource, cls):
                continue
            # Filter by class_id
            if format is not None and resource.metadata.format != format:
                continue
            # Filter by workflow state
            if state is not None and resource.get_workflow_state() != state:
                continue
            # All filters passed
            yield resource


    def get_human_size(self):
        names = self.get_names()
        size = len(names)

        return MSG(u'$n obs').gettext(n=size)


    #######################################################################
    # User interface
    #######################################################################
    def get_view(self, name, query=None):
        # Add resource form
        if name == 'new_resource':
            if query is not None and 'type' in query:
                view = get_resource_class(query['type']).new_instance
                if isinstance(view, BaseView):
                    return view

        # Default
        return DBResource.get_view(self, name, query)


    #######################################################################
    # Views
    view = Folder_View()
    new_resource = Folder_NewResource()
    browse_content = Folder_BrowseContent()
    rename = Folder_Rename()
    preview_content = Folder_PreviewContent()
    last_changes = Folder_LastChanges()
    orphans = Folder_Orphans()
    thumb = Folder_Thumbnail()



###########################################################################
# Register
###########################################################################
register_resource_class(Folder)
register_resource_class(Folder, format="application/x-not-regular-file")



# -*- coding: UTF-8 -*-
# Copyright (C) 2006-2007 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007 Henry Obein <henry@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


resources_registry = {}

def register_resource_class(resource_class, format=None):
    if format is None:
        format = resource_class.class_id
    resources_registry[format] = resource_class


def get_resource_class(class_id, is_file=True):
    if class_id in resources_registry:
        return resources_registry[class_id]

    if '/' in class_id:
        class_id = class_id.split('/')[0]
        if class_id in resources_registry:
            return resources_registry[class_id]

    # Default
    if is_file:
        return resources_registry['application/octet-stream']

    return resources_registry['application/x-not-regular-file']


websites_registry = {}

def register_website(website):
    websites_registry[website.class_id] = website


def get_register_websites():
    return websites_registry.itervalues()


def get_website_class(class_id):
    return websites_registry[class_id]



# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2006-2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007-2008 Sylvain Taverne <sylvain@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from copy import deepcopy
from string import Template

# Import from itools
from itools.datatypes import Email, String, Unicode
from itools.gettext import MSG
from itools.uri import Path
from itools.xapian import TextField, KeywordField
from itools.web import INFO

# Import from ikaaro
from access import AccessControl
from datatypes import Password
from folder import Folder
from folder_views import Folder_BrowseContent
from registry import register_resource_class, get_resource_class
from resource_views import DBResource_Edit
from user_views import User_ConfirmRegistration, User_EditAccount
from user_views import User_EditPassword, User_EditPreferences, User_Profile
from user_views import User_ResendConfirmation, User_Tasks
from user_views import User_ChangePasswordForgotten
from utils import crypt_password, generate_password
from views import MessageView



class User(AccessControl, Folder):

    class_id = 'user'
    class_version = '20081217'
    class_title = MSG(u'User')
    class_icon16 = 'icons/16x16/user.png'
    class_icon48 = 'icons/48x48/user.png'
    class_views = ['profile', 'edit_account', 'edit_preferences',
                   'edit_password', 'tasks']


    ########################################################################
    # Metadata
    ########################################################################
    @classmethod
    def get_metadata_schema(cls):
        return {
            'firstname': Unicode,
            'lastname': Unicode,
            'email': Email,
            'password': Password,
            'user_language': String,
            'user_must_confirm': String,
            # Backwards compatibility
            'username': String,
        }


    ########################################################################
    # Indexing
    ########################################################################
    def get_catalog_fields(self):
        fields = Folder.get_catalog_fields(self)
        fields += [KeywordField('email', is_stored=True),
                   TextField('lastname', is_stored=True),
                   TextField('firstname', is_stored=True),
                   # Login Name
                   KeywordField('username', is_stored=True)]
        return fields


    def get_catalog_values(self):
        values = Folder.get_catalog_values(self)
        values['email'] = self.get_property('email')
        values['username'] = self.get_login_name()
        values['firstname'] = self.get_property('firstname')
        values['lastname'] = self.get_property('lastname')
        return values


    def get_canonical_path(self):
        return Path('/users/%s' % self.name)


    ########################################################################
    # API
    ########################################################################
    def get_title(self, language=None):
        firstname = self.get_property('firstname')
        lastname = self.get_property('lastname')
        if firstname:
            if lastname:
                return '%s %s' % (firstname, lastname)
            return firstname
        if lastname:
            return lastname
        return self.get_login_name()


    def get_login_name(self):
        # FIXME Check first the username (for compatibility with 0.14)
        username = self.get_property('username')
        if username:
            return username
        return self.get_property('email')


    def set_password(self, password):
        crypted = crypt_password(password)
        self.set_property('password', crypted)


    def authenticate(self, password):
        # Is password crypted?
        if password == self.get_property('password'):
            return True
        # Is password clear?
        crypted = crypt_password(password)
        return crypted == self.get_property('password')


    def get_groups(self):
        """Returns all the role aware handlers where this user is a member.
        """
        root = self.get_root()
        if root is None:
            return ()

        results = root.search(is_role_aware=True, members=self.name)
        groups = [ x.abspath for x in results.get_documents() ]
        return tuple(groups)


    def set_auth_cookie(self, context, password):
        username = str(self.name)
        crypted = crypt_password(password)
        cookie = Password.encode('%s:%s' % (username, crypted))
        expires = context.request.get_parameter('iAuthExpires')
        if expires is None:
            context.set_cookie('__ac', cookie, path='/')
        else:
            context.set_cookie('__ac', cookie, path='/', expires=expires)


    ########################################################################
    # Email: Register confirmation & Password forgotten
    ########################################################################

    confirmation_subject = MSG(u"Confirmation required")
    confirmation_txt = MSG(u"To confirm your identity, click the link:"
                           u"\n"
                           u"\n $uri")
    def send_confirmation(self, context, email):
        self.send_confirm_url(context, email, self.confirmation_subject,
            self.confirmation_txt, ';confirm_registration')


    forgotten_subject = MSG(u"Choose a new password")
    forgotten_txt = MSG(u"To choose a new password, click the link:"
                        u"\n"
                        u"\n $uri")
    def send_forgotten_password(self, context, email):
        self.send_confirm_url(context, email, self.forgotten_subject,
            self.forgotten_txt, ';change_password_forgotten')


    def send_confirm_url(self, context, email, subject, text, view):
        # Set the confirmation key
        if self.has_property('user_must_confirm'):
            key = self.get_property('user_must_confirm')
        else:
            key = generate_password(30)
            self.set_property('user_must_confirm', key)

        # Build the confirmation link
        confirm_url = deepcopy(context.uri)
        path = '/users/%s/%s' % (self.name, view)
        confirm_url.path = Path(path)
        confirm_url.query = {'key': key, 'username': self.get_login_name()}
        confirm_url = str(confirm_url)
        text = text.gettext(uri=confirm_url)
        context.root.send_email(email, subject.gettext(), text=text)

    ########################################################################
    # Access control
    def is_self_or_admin(self, user, resource):
        # You are nobody here, ha ha ha
        if user is None:
            return False

        # In my home I am the king
        if self.name == user.name:
            return True

        # The all-powerfull
        return self.is_admin(user, resource)


    is_allowed_to_edit = is_self_or_admin


    #######################################################################
    # Views
    #######################################################################
    resend_confirmation = User_ResendConfirmation()
    confirm_registration = User_ConfirmRegistration()
    change_password_forgotten = User_ChangePasswordForgotten()
    profile = User_Profile()
    edit_account = User_EditAccount()
    edit_preferences = User_EditPreferences()
    edit_password = User_EditPassword()
    tasks = User_Tasks()


    def update_20081217(self):
        # Some users were registered with multilingual names
        for key in ('firstname', 'lastname'):
            value, language = self.get_property_and_language(key)
            if language is not None:
                self.del_property(key)
                self.set_property(key, value)



class UserFolder(Folder):

    class_id = 'users'
    class_version = '20071215'
    class_icon16 = 'icons/16x16/userfolder.png'
    class_icon48 = 'icons/48x48/userfolder.png'
    class_views = ['view', 'browse_content', 'edit']


    def get_document_types(self):
        return [get_resource_class('user')]


    #######################################################################
    # API
    #######################################################################
    def set_user(self, email=None, password=None):
        # Calculate the user id
        ids = []
        for key in self.get_names():
            try:
                key = int(key)
            except ValueError:
                continue
            ids.append(key)
        if ids:
            ids.sort()
            user_id = str(ids[-1] + 1)
        else:
            user_id = '0'

        # Add the user
        cls = get_resource_class('user')
        user = cls.make_resource(cls, self, user_id)
        # Set the email and paswword
        if email is not None:
            user.set_property('email', email)
        if password is not None:
            user.set_password(password)

        # Return the user
        return user


    def get_usernames(self):
        """Return all user names."""
        names = self._get_names()
        return frozenset(names)


    #######################################################################
    # Back-Office
    #######################################################################
    browse_content = Folder_BrowseContent(access='is_admin')
    edit = DBResource_Edit(access='is_admin')


    #######################################################################
    # View
    view = MessageView(
        access='is_admin',
        title=MSG(u'View'),
        icon='view.png',
        message=INFO(u'To manage the users please go '
                     u'<a href="/;browse_users">here</a>.'))



###########################################################################
# Register
###########################################################################
register_resource_class(UserFolder)
register_resource_class(User)



# -*- coding: UTF-8 -*-
# Copyright (C) 2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from itools
from itools.gettext import MSG

# Import from ikaaro
from datatypes import CopyCookie
import messages


class Button(object):

    access = False
    confirm = None
    css = 'button_ok'
    name = None
    title = None

    def __init__(self, **kw):
        for key in kw:
            setattr(self, key, kw[key])


    @classmethod
    def show(cls, resource, context, items):
        if len(items) == 0:
            return False
        ac = resource.get_access_control()
        return ac.is_access_allowed(context.user, resource, cls)



class RemoveButton(Button):

    access = 'is_allowed_to_remove'
    confirm = messages.MSG_DELETE_SELECTION
    css = 'button_delete'
    name = 'remove'
    title = MSG(u'Remove')



class RenameButton(Button):

    access = 'is_allowed_to_move'
    css = 'button_rename'
    name = 'rename'
    title = MSG(u'Rename')



class CopyButton(Button):

    access = 'is_allowed_to_copy'
    css = 'button_copy'
    name = 'copy'
    title = MSG(u'Copy')



class CutButton(Button):

    access = 'is_allowed_to_move'
    css = 'button_cut'
    name = 'cut'
    title = MSG(u'Cut')



class PasteButton(Button):

    access = 'is_allowed_to_move'
    css = 'button_paste'
    name = 'paste'
    title = MSG(u'Paste')


    @classmethod
    def show(cls, resource, context, items):
        cut, paths = context.get_cookie('ikaaro_cp', type=CopyCookie)
        if len(paths) == 0:
            return False
        ac = resource.get_access_control()
        return ac.is_access_allowed(context.user, resource, cls)



class PublishButton(Button):

    access = 'is_allowed_to_publish'
    css = 'button_publish'
    name = 'publish'
    title = MSG(u'Publish')


    @classmethod
    def show(cls, resource, context, items):
        ac = resource.get_access_control()
        for item in items:
            if ac.is_allowed_to_trans(context.user, item, 'publish'):
                return True
        return False



class OrderUpButton(Button):

    access = 'is_allowed_to_edit'
    name = 'order_up'
    title = MSG(u'Order up')



class OrderDownButton(Button):

    access = 'is_allowed_to_edit'
    name = 'order_down'
    title = MSG(u'Order down')



class OrderTopButton(Button):

    access = 'is_allowed_to_edit'
    name = 'order_top'
    title = MSG(u'Order top')



class OrderBottomButton(Button):

    access = 'is_allowed_to_edit'
    name = 'order_bottom'
    title = MSG(u'Order bottom')




# -*- coding: UTF-8 -*-
# Copyright (C) 2006-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2006-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2008 David Versmisse <david.versmisse@itaapy.com>
# Copyright (C) 2008 Henry Obein <henry@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from os import fdopen
import sys
from tempfile import mkstemp

# Import from xapian
from xapian import DatabaseOpeningError

# Import from itools
from itools.datatypes import Boolean, Integer, String, Tokens
from itools.handlers import ConfigFile, SafeDatabase
from itools.uri import get_absolute_reference2
from itools import vfs
from itools.web import Server as BaseServer
from itools.xapian import Catalog

# Import from ikaaro
from folder import Folder
from metadata import Metadata
from registry import get_resource_class
from utils import is_pid_running
from versioning import VersioningAware
from website import WebSite



class ServerConfig(ConfigFile):

    schema = {
        'modules': Tokens(default=()),
        'listen-address': String(default=''),
        'listen-port': Integer(default=8080),
        'smtp-host': String(default=''),
        'smtp-from': String(default=''),
        'smtp-login': String(default=''),
        'smtp-password': String(default=''),
        'debug': Boolean(default=False),
    }




def ask_confirmation(message, confirm=False):
    if confirm is True:
        print message + 'Y'
        return True

    sys.stdout.write(message)
    sys.stdout.flush()
    line = sys.stdin.readline()
    line = line.strip().lower()
    return line == 'y'



def get_config(target):
    return ServerConfig('%s/config.conf' % target)



def load_modules(config):
    """Load Python packages and modules.
    """
    modules = config.get_value('modules')
    for name in modules:
        name = name.strip()
        exec('import %s' % name)



def get_pid(target):
    try:
        pid = open('%s/pid' % target).read()
    except IOError:
        return None

    pid = int(pid)
    if is_pid_running(pid):
        return pid
    return None



def get_root(database, target):
    path = '%s/database/.metadata' % target
    metadata = database.get_handler(path, cls=Metadata)
    cls = get_resource_class(metadata.format)
    # Build the root resource
    root = cls(metadata)
    # FIXME Should be None
    root.name = root.class_title.message.encode('utf_8')
    return root



class Server(BaseServer):

    def __init__(self, target, address=None, port=None, debug=False,
                 read_only=False):
        target = get_absolute_reference2(target)
        self.target = target
        path = target.path

        # Load the config
        config = get_config(target)
        load_modules(config)

        # Find out the IP to listen to
        if not address:
            address = config.get_value('listen-address').strip()

        # Find out the port to listen
        if not port:
            port = config.get_value('listen-port')

        # Contact Email
        self.smtp_from = config.get_value('smtp-from')

        # Full-text indexing
        self.index_text =  config.get_value('index-text', type=Boolean,
                                            default=True)

        # Logs
        event_log = '%s/log/events' % path
        access_log = '%s/log/access' % path
        debug = debug or config.get_value('debug')

        # The database
        if debug:
            database = SafeDatabase('%s/database.commit' % path, event_log)
        else:
            database = SafeDatabase('%s/database.commit' % path)
        self.database = database
        # The catalog
        # FIXME Backwards compatibility with 0.20
        try:
            self.catalog = Catalog('%s/catalog' % target, read_only=read_only)
        except DatabaseOpeningError, e:
            print e
            self.catalog = None

        # Find out the root class
        root = get_root(database, target)

        # Events
        self.resources_added = set()
        self.resources_changed = set()
        self.resources_removed = set()

        # Initialize
        BaseServer.__init__(self, root, address=address, port=port,
                            access_log=access_log, event_log=event_log,
                            debug=debug, pid_file='%s/pid' % path)


    #######################################################################
    # API / Private
    #######################################################################
    def get_pid(self):
        return get_pid(self.target.path)


    def send_email(self, message):
        # Check the SMTP host is defined
        config = get_config(self.target)
        if not config.get_value('smtp-host'):
            raise ValueError, '"smtp-host" is not set in config.conf'

        spool = self.target.resolve2('spool')
        spool = str(spool.path)
        tmp_file, tmp_path = mkstemp(dir=spool)
        file = fdopen(tmp_file, 'w')
        try:
            file.write(message.as_string())
        finally:
            file.close()


    def get_databases(self):
        return [self.database, self.catalog]


    def abort_transaction(self, context):
        # Clear events
        self.resources_removed.clear()
        self.resources_added.clear()
        self.resources_changed.clear()
        # Follow-up
        BaseServer.abort_transaction(self, context)


    def before_commit(self):
        root = self.root
        catalog = self.catalog
        # Removed
        for path in self.resources_removed:
            catalog.unindex_document(path)
        self.resources_removed.clear()

        # Added
        for path in self.resources_added:
            resource = root.get_resource(path)
            if isinstance(resource, VersioningAware):
                resource.commit_revision()
            catalog.index_document(resource)
        self.resources_added.clear()

        # Changed
        for path in self.resources_changed:
            resource = root.get_resource(path)
            if isinstance(resource, VersioningAware):
                resource.commit_revision()
            catalog.unindex_document(path)
            catalog.index_document(resource)
        self.resources_changed.clear()


    #######################################################################
    # API / Public
    #######################################################################
    def init_context(self, context):
        BaseServer.init_context(self, context)
        # Set the list of needed resources. The method we are going to
        # call may need external resources to be rendered properly, for
        # example it could need an style sheet or a javascript file to
        # be included in the html head (which it can not control). This
        # attribute lets the interface to add those resources.
        context.styles = []
        context.scripts = []
        context.message = None


    def find_site_root(self, context):
        # Default to root
        root = self.root
        context.site_root = root

        # Check we have a URI
        uri = context.uri
        if uri is None:
            return

        # The site root depends on the host
        hostname = uri.authority.host

        # Check first the root
        if hostname in root.get_property('vhosts'):
            return

        # Check the sub-sites
        for site in root.search_resources(cls=WebSite):
            if hostname in site.get_property('vhosts'):
                context.site_root = site
                return


    def remove_resource(self, resource):
        resources_removed = self.resources_removed
        resources_added = self.resources_added
        resources_changed = self.resources_changed

        if isinstance(resource, Folder):
            for x in resource.traverse_resources():
                path = str(x.get_canonical_path())
                if path in resources_added:
                    resources_added.remove(path)
                if path in resources_changed:
                    resources_changed.remove(path)
                resources_removed.add(path)
        else:
            path = str(resource.get_canonical_path())
            if path in resources_added:
                resources_added.remove(path)
            if path in resources_changed:
                resources_changed.remove(path)
            resources_removed.add(path)


    def add_resource(self, resource):
        if isinstance(resource, Folder):
            for x in resource.traverse_resources():
                path = str(x.get_canonical_path())
                self.resources_added.add(path)
        else:
            path = str(resource.get_canonical_path())
            self.resources_added.add(path)


    def change_resource(self, resource):
        path = str(resource.get_canonical_path())
        if path in self.resources_removed:
            raise ValueError, 'XXX'
        if path in self.resources_added:
            return
        self.resources_changed.add(path)




# -*- coding: UTF-8 -*-
# Copyright (C) 2006-2007 Nicolas Deram <nicolas@itaapy.com>
# Copyright (C) 2006-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2008 Sylvain Taverne <sylvain@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from os import getenv, listdir

# Import from itools
from itools import get_abspath
from itools.gettext import register_domain
from itools.utils import get_version

# Import from ikaaro
from calendar_ import CalendarTable
from file import File
from folder import Folder
from html import WebPage
import root


# The version
__version__ = get_version()


# Register the itools domain
path = get_abspath('locale')
register_domain('ikaaro', path)

# Register document types
Folder.register_document_type(WebPage)
Folder.register_document_type(Folder)
Folder.register_document_type(File)
Folder.register_document_type(CalendarTable)

# Import ikaaro sub-packages (NOTE must be imported after so they are
# register after)
import forum
import tracker
try:
    import docutils
except ImportError:
    print "docutils is not installed, wiki deactivated."
else:
    import wiki


###########################################################################
# Check for required software
###########################################################################
cmds = ['wvText', 'xlhtml', 'ppthtml', 'pdftotext', 'unrtf']

paths = getenv('PATH').split(':')
all_names = set()
for path in paths:
    path = path.strip()
    try:
        names = listdir(path)
    except OSError:
        pass
    else:
        all_names = all_names.union(names)

for name in cmds:
    if name not in all_names:
        print 'You need to install the command "%s".' % name



# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007 Nicolas Deram <nicolas@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from random import sample
from hashlib import sha1
from sys import platform
from urllib import quote

# Import from itools
from itools.utils import freeze
from itools.web import get_context
from itools.xapian import AllQuery, PhraseQuery, NotQuery, OrQuery, StartQuery
from itools.xapian import AndQuery

if platform[:3] == 'win':
    from utils_win import is_pid_running, kill
else:
    from utils_unix import is_pid_running, kill


###########################################################################
# Navigation helper functions
###########################################################################
def get_parameters(prefix, **kw):
    """Gets the parameters from the request form, the keyword argument
    specifies which are the parameters to get and which are their default
    values.

    The prefix argument lets to create different namespaces for the
    parameters, so the same page web can have different sections with
    different but equivalent parameters.

    For example, call it like:

      get_parameters('resources', sortby='id', sortorder='up')
    """
    # Get the form field from the request (a zope idiom)
    get_parameter = get_context().request.get_parameter

    # Get the parameters
    parameters = {}
    for key, value in kw.items():
        parameters[key] = get_parameter('%s_%s' % (prefix, key),
                                        default=value)

    return parameters


def preserve_parameters(preserve=freeze([])):
    """Returns an HTML snippet with hidden input html elements, there will
    be one element for each request parameter that starts with any of
    the prefixes contained in the preserve parameter.

    This lets to pass url request parameters through form actions, so we
    don't lose important navigation information.
    """
    snippet = []

    form = get_context().request.get_form()
    for k, v in form.items():
        for prefix in preserve:
            if k.startswith(prefix):
                snippet.append('<input type="hidden" name="%s" value="%s">'
                               % (k, quote(v)))
                break

    return '\n'.join(snippet)



###########################################################################
# Languages
###########################################################################

# Mark for translatios
u'Basque'
u'Catalan'
u'English'
u'French'
u'German'
u'Hungarian'
u'Italian'
u'Japanese'
u'Portuguese'
u'Spanish'


###########################################################################
# String format for display
###########################################################################

def reduce_string(title='', word_treshold=15, phrase_treshold=40):
    """Reduce words and string size.
    """
    ellipsis = '…'
    if isinstance(title, unicode):
        ellipsis = u'…'
    words = title.strip().split(' ')
    for i, word in enumerate(words):
        if len(word) > word_treshold:
            words.pop(i)
            word = word[:word_treshold] + ellipsis
            words.insert(i, word)
    title = ' '.join(words)
    if len(title) > phrase_treshold:
        title = title[:phrase_treshold] + ellipsis
    return title


###########################################################################
# User and Authentication
###########################################################################
# ASCII letters and digits, except the characters: 0, O, 1, l
tokens = 'abcdefghijkmnopqrstuvwxyzABCDEFGHIJKLMNPQRSTUVWXYZ23456789'
def generate_password(length=6):
    return ''.join(sample(tokens, length))


def crypt_password(password):
    return sha1(password).digest()


###########################################################################
# Generate next name
###########################################################################
def generate_name(name, used, suffix='_'):
    """Generate a name which is not in list "used" based on name and suffix.
    Example:
      With name='toto.txt', used=['toto.txt', 'toto_0.txt']
      --> toto.txt and toto_0.txt are used so it returns toto_1.txt
      With name='toto.txt', used=['toto.txt', 'toto_0.txt'], suffix='_copy_'
      --> toto.txt is used so it returns toto_copy_0.txt
    """
    if name not in used:
        return name

    items = name.split('.', 1)
    basename = items[0]
    extent = ''
    if len(items) > 1:
        extent = '.%s' % items[1]

    # 1st time called
    if suffix not in basename:
        index = 0
    else:
        basename, index = basename.rsplit(suffix, 1)
        try:
            index = int(index) + 1
        except ValueError:
            basename = '%s%s%s' % (basename, suffix, index)
            index = 0

    name = ''.join([basename, suffix, str(index), extent])
    while name in used:
        index += 1
        name = ''.join([basename, suffix, str(index), extent])

    return str(name)



###########################################################################
# Index and Search
###########################################################################
def get_base_path_query(abspath, include_container=False):
    """Builds a query that will return all the objects within the given
    absolute path, like it is returned by 'resource.get_abspath()'.

    If 'include_container' is true the resource at the given path will be
    returned too.
    """
    # Case 1: everything
    if abspath == '/' and include_container is True:
        return AllQuery()

    # Case 2: everything but the root
    if abspath == '/':
        return NotQuery(PhraseQuery('abspath', '/'))

    # Case 3: some subfolder
    content = StartQuery('abspath', abspath + '/')
    if include_container is False:
        return content

    container = PhraseQuery('abspath', abspath)
    return OrQuery(container, content)




# -*- coding: UTF-8 -*-
# Copyright (C) 2006-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2006-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Henry Obein <henry@itaapy.com>
# Copyright (C) 2008 Nicolas Deram <nicolas@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from operator import itemgetter
from warnings import warn

# Import from itools
from itools.gettext import MSG
from itools.html import xhtml_uri
from itools.stl import set_prefix
from itools.uri import get_reference
from itools.xml import XMLParser, START_ELEMENT

# Import from ikaaro
from ikaaro.file import Image
from ikaaro.folder import Folder
from ikaaro.html import ResourceWithHTML, WebPage
from ikaaro.registry import register_resource_class
from ikaaro.views import CompositeForm, ContextMenu
from ikaaro.wiki import WikiPage
from ikaaro.workflow import WorkflowAware


class Dressable_Menu(ContextMenu):

    title = MSG(u'Edit')

    def get_items(self, resource, context):
        items = []
        base_path = ''
        if not isinstance(resource, Dressable):
            resource = resource.parent
            base_path = '../'
            if not isinstance(resource, Dressable):
                return []

        for name, value in resource.layout.iteritems():
            if not isinstance(value, tuple):
                msg = u'Layout items MUST be 2 values tuples: '
                msg += u'"%s" is incorrect' % name
                warn(msg)
                continue
            name, cls = value
            if resource.has_resource(name):
                # Add edit link
                items.append(
                    {'name': name,
                     'title': MSG(u'%s %s' % (cls.class_title.gettext(), name)),
                     'href': '%s%s/;edit' % (base_path, name),
                     'class': 'nav_active'})
            else:
                # Add new_resource link
                items.append(
                    {'name': name,
                     'title': MSG(u'Add new %s' % cls.class_title.gettext()),
                     'href': ('%s;new_resource?type=%s&title=%s' %
                              (base_path, cls.class_id, name)),
                     'class': 'nav_active'})
        items.sort(key=itemgetter('name'))
        # Dressable metadata
        items.insert(0,
            {'title': MSG(u'Metadata'),
             'href': '%s;edit' % base_path,
             'class': 'nav_active'})
        # Back to preview
        items.append(
            {'title': MSG(u'Preview'),
             'href': '%s.' % base_path,
             'class': 'nav_active'})
        return items



class Dressable_View(CompositeForm):

    access = 'is_allowed_to_view'
    title = MSG(u'View')
    icon = 'view.png'
    template = '/ui/future/dressable_view.xml'


    def GET(self, resource, context):
        stream = CompositeForm.GET(self, resource, context)
        prefix = '%s/' % resource.name
        return set_prefix(stream, prefix)


    def get_view(self, resource, context, item):
        if isinstance(item, WebPage):
            return item.view.GET
        if isinstance(item, Image):
            return resource._get_image
        # TODO We should implement a specific view to display wiki page as HTML
        if isinstance(item, WikiPage):
            return item.view.GET
        return getattr(self, item, None)


    def get_namespace(self, resource, context):
        items = []
        for name, value in resource.layout.iteritems():
            content = ''
            if isinstance(value, tuple):
                name, kk = value
                if resource.has_resource(name):
                    item = resource.get_resource(name)
                    # get view to show current item
                    method = self.get_view(resource, context, item)
                    value = method(item, context)
                else:
                    value = None
            else:
                # get view to display
                value = getattr(resource, 'data')(context)
            items.append({'id': name, 'content': value})
        return {'items': items}



class Dressable(Folder, ResourceWithHTML):
    """A Dressable resource is a folder with a specific view which is defined
    by the layout. In addition of the layout, it is necessary to redefine
    the variable __fixed_handlers__.
    """

    class_id = 'dressable'
    class_version = '20081118'
    class_title = MSG(u'Dressable')
    class_description = MSG(u'A dressable folder')
    class_views = ['view', 'browse_content', 'preview_content']
    __fixed_handlers__ = ['index']
    layout = {'content': ('index', WebPage),
              'image1': ('image1', Image)}
    context_menus = [Dressable_Menu()]


    @staticmethod
    def _make_resource(cls, folder, name, **kw):
        Folder._make_resource(cls, folder, name, **kw)
        # populate the dressable
        cls._populate(cls, folder, name)


    @staticmethod
    def _populate(cls, folder, base_name):
        """Populate the dressable from the layout"""
        for key, data in cls.layout.iteritems():
            if isinstance(data, tuple):
                handler_name, handler_cls = data
                if issubclass(handler_cls, WebPage):
                    full_name = '%s/%s.metadata' % (base_name, handler_name)
                    metadata = handler_cls.build_metadata()
                    if issubclass(handler_cls, WorkflowAware):
                        metadata.set_property('state', 'public')
                    folder.set_handler(full_name, metadata)


    def get_document_types(self):
        return [self.__class__] + Folder.get_document_types(self)


    def get_html_document(self, language=None):
        resource = self.get_resource('index')
        return resource.get_html_document(language)


    def _get_image(self, item, context):
        return XMLParser('<img src="%s/;download"/>' % context.get_link(item))


    #######################################################################
    # User interface
    #######################################################################
    view = Dressable_View()


    #######################################################################
    # Update
    #######################################################################
    def update_20081118(self):
        def fix_links(stream):
            for event in stream:
                type, value, line = event
                if type != START_ELEMENT:
                    yield event
                    continue
                tag_uri, tag_name, attributes = value
                if tag_uri != xhtml_uri:
                    yield event
                    continue
                if tag_name not in ('img', 'a'):
                    yield event
                    continue
                if tag_name == 'img':
                    attr_name = 'src'
                else:
                    attr_name = 'href'
                value = attributes.get((None, attr_name))
                if value is None:
                    yield event
                    continue
                uri = get_reference(value)
                if uri.scheme or uri.authority or not uri.path:
                    yield event
                    continue
                if value.startswith('/ui/'):
                    yield event
                    continue
                # Fix link
                uri = '../' + str(uri)
                attributes = attributes.copy()
                attributes[(None, attr_name)] = str(uri)
                yield START_ELEMENT, (tag_uri, tag_name, attributes), line

        languages = self.get_site_root().get_property('website_languages')
        for wp in self.search_resources(cls=WebPage):
            for language in languages:
                handler = wp.get_handler(language=language)
                events = list(fix_links(handler.events))
                handler.set_changed()
                handler.events = events


register_resource_class(Dressable)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Copyright (C) 2006 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2006-2007 Juan David Ibáñez Palomar <jdavid@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from optparse import OptionParser

# Import from itools
import itools
from itools import vfs
from itools.handlers import (SafeDatabase, READY, TRANSACTION_PHASE1,
    TRANSACTION_PHASE2)

# Import from ikaaro
from ikaaro.server import ask_confirmation


def restore(parser, options, target):
    database = SafeDatabase('%s/database.commit' % target)

    state = database.get_state()
    if state == READY:
        print 'Everything seems fine.'
    elif state == TRANSACTION_PHASE1:
        msg = 'The latest transaction failed. Clean up database (y/N)? '
        if ask_confirmation(msg) is True:
            print '  * Cleaning...'
            database.rollback()
            print 'DONE'
    elif state == TRANSACTION_PHASE2:
        msg = 'The latest transaction was not saved. Finish now (y/N)? '
        if ask_confirmation(msg) is True:
            print '  * Saving transaction...'
            database.save_changes_forever()
            print 'DONE'
    else:
        raise ValueError, 'unknown database state'



if __name__ == '__main__':
    # The command line parser
    usage = '%prog TARGET'
    version = 'itools %s' % itools.__version__
    description = ('Restore the TARGET ikaaro instance if broken. To be'
                   ' used after a crash.')
    parser = OptionParser(usage, version=version, description=description)

    options, args = parser.parse_args()
    if len(args) != 1:
        parser.error('incorrect number of arguments')

    target = args[0]

    # Action!
    restore(parser, options, target)



# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Luis Arturo Belmar-Letelier <luis@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007-2008 Nicolas Deram <nicolas@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from itools
from itools.datatypes import Boolean, Integer, String, Unicode
from itools.gettext import MSG
from itools.handlers import ConfigFile

# Import from ikaaro
from ikaaro.registry import register_resource_class
from ikaaro.text import Text



class StoredSearchFile(ConfigFile):

    schema = {
        'text': Unicode,
        'mtime': Integer(default=0),
        'product': Integer(multiple=True),
        'module': Integer(multiple=True),
        'version': Integer(multiple=True),
        'type': Integer(multiple=True),
        'priority': Integer(multiple=True),
        'assigned_to': String(multiple=True),
        'state': Integer(multiple=True),
        'sort_by': String,
        'reverse': Boolean,
        }


class StoredSearch(Text):

    class_id = 'stored_search'
    class_version = '20071215'
    class_title = MSG(u'Stored Search')
    class_handler = StoredSearchFile


    def get_values(self, name, type=None):
        return self.handler.get_value(name)

    def set_values(self, name, value, type):
        if isinstance(value, list):
            value = [ type.encode(x) for x in value ]
            value = ' '.join(value)
        self.handler.set_value(name, value)




# Register
register_resource_class(StoredSearch)



# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007-2008 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007-2008 Nicolas Deram <nicolas@itaapy.com>
# Copyright (C) 2007-2008 Sylvain Taverne <sylvain@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from itools
from itools.csv import UniqueError, Property, is_multilingual
from itools.datatypes import Integer, Enumerate, Tokens
from itools.gettext import MSG
from itools.web import INFO, ERROR
from itools.xapian import PhraseQuery

# Import from ikaaro
from buttons import RemoveButton, OrderUpButton, OrderDownButton
from buttons import OrderBottomButton, OrderTopButton
from forms import AutoForm
import messages
from resource_views import EditLanguageMenu
from views import SearchForm



class Table_View(SearchForm):

    access = 'is_allowed_to_view'
    access_POST = 'is_allowed_to_edit'
    title = MSG(u'View')
    icon = 'view.png'

    schema = {
        'ids': Integer(multiple=True, mandatory=True),
    }

    def get_widgets(self, resource, context):
        return resource.get_form()


    def get_search_schema(self, resource, context):
        return resource.handler.record_schema


    def get_items(self, resource, context):
        search_query = None

        # Build the search query
        if self.search_template is not None:
            query = context.query
            search_term = query['search_term'].strip()
            if search_term:
                search_field = query['search_field']
                search_query = PhraseQuery(search_field, search_term)

        # Ok
        items = resource.handler.search(search_query)
        return list(items)


    def get_search_fields(self, resource, context):
        search_fields = []
        schema = self.get_search_schema(resource, context)
        for widget in self.get_widgets(resource, context):
            if hasattr(schema[widget.name], 'index'):
                title = getattr(widget, 'title', widget.name)
                search_fields.append((widget.name, title))
        return search_fields


    def sort_and_batch(self, resource, context, items):
        # Sort
        sort_by = context.query['sort_by']
        reverse = context.query['reverse']
        if sort_by:
            items.sort(cmp=lambda x,y: cmp(getattr(x, sort_by),
                       getattr(y, sort_by)), reverse=reverse)

        # Batch
        start = context.query['batch_start']
        size = context.query['batch_size']
        return items[start:start+size]


    def get_table_columns(self, resource, context):
        columns = [
            ('checkbox', None),
            ('id', MSG(u'id'))]
        # From the schema
        for widget in self.get_widgets(resource, context):
            column = (widget.name, getattr(widget, 'title', widget.name))
            columns.append(column)

        return columns


    def get_item_value(self, resource, context, item, column):
        if column == 'checkbox':
            return item.id, False
        elif column == 'id':
            id = item.id
            link = context.get_link(resource)
            return id, '%s/;edit_record?id=%s' % (link, id)

        # Columns
        handler = resource.handler
        value = handler.get_record_value(item, column)
        datatype = handler.get_record_datatype(column)

        # Multilingual
        if is_multilingual(datatype):
            return value

        # Multiple
        is_multiple = datatype.multiple
        is_tokens = issubclass(datatype, Tokens)

        if is_multiple or is_tokens:
            if is_multiple:
                value.sort()
            value_length = len(value)
            if value_length > 0:
                rmultiple = value_length > 1
                value = value[0]
            else:
                rmultiple = False
                value = None

        # Enumerate
        is_enumerate = getattr(datatype, 'is_enumerate', False)
        if is_enumerate:
            value = datatype.get_value(value)

        if is_multiple or is_tokens:
            return value, rmultiple
        return value


    table_actions = [RemoveButton]


    #######################################################################
    # Form Actions
    #######################################################################
    def action_remove(self, resource, context, form):
        ids = form['ids']
        for id in ids:
            resource.handler.del_record(id)
        # Reindex the resource
        context.server.change_resource(resource)

        context.message = INFO(u'Record deleted.')



###########################################################################
# Add/Edit records
###########################################################################
class Table_AddEditRecord(AutoForm):

    access = 'is_allowed_to_edit'
    context_menus = [EditLanguageMenu()]


    def get_schema(self, resource, context):
        return resource.get_schema()


    def get_widgets(self, resource, context):
        return resource.get_form()


    def action(self, resource, context, form):
        """Code shared by the add & edit actions.  It builds a new record
        from the form.
        """
        schema = self.get_schema(resource, context)
        language = resource.get_content_language(context)

        # Builds a new record from the form.
        record = {}
        for name in schema:
            datatype = schema[name]
            value = form[name]
            if is_multilingual(datatype):
                value = value[0]
                value = Property(value, language=language)
            elif datatype.multiple:
                # textarea -> string
                if not issubclass(datatype, Enumerate):
                    value = [ x.strip() for x in value.splitlines() ]
                    value = [ datatype.decode(x) for x in value if x ]
            record[name] = value

        # Change
        try:
            self.action_add_or_edit(resource, context, record)
        except UniqueError, error:
            title = resource.get_field_title(error.name)
            context.message = ERROR(error, field=title, value=error.value)
        except ValueError, error:
            message = ERROR(u'Error: $msg', msg=str(error))
            context.message = message
        else:
            return self.action_on_success(resource, context)



class Table_AddRecord(Table_AddEditRecord):

    title = MSG(u'Add Record')
    icon = 'new.png'
    submit_value = MSG(u'Add')


    def action_add_or_edit(self, resource, context, record):
        resource.handler.add_record(record)
        # Reindex the resource
        context.server.change_resource(resource)


    def action_on_success(self, resource, context):
        n = len(resource.handler.records) - 1
        goto = ';edit_record?id=%s' % n
        return context.come_back(MSG(u'New record added.'), goto=goto)



class Table_EditRecord(Table_AddEditRecord):

    title = MSG(u'Edit record ${id}')
    query_schema = {'id': Integer(mandatory=True)}


    def get_value(self, resource, context, name, datatype):
        handler = resource.get_handler()
        # Get the record
        id = context.query['id']
        record = handler.get_record(id)
        # Is mulitilingual
        if is_multilingual(datatype):
            language = resource.get_content_language(context)
            return handler.get_record_value(record, name, language=language)

        return handler.get_record_value(record, name)


    def get_title(self, context):
        id = context.query['id']
        return self.title.gettext(id=id)


    def action_add_or_edit(self, resource, context, record):
        id = context.query['id']
        resource.handler.update_record(id, **record)
        # Reindex the resource
        context.server.change_resource(resource)


    def action_on_success(self, resource, context):
        context.message = messages.MSG_CHANGES_SAVED



###########################################################################
# Oredered Views
###########################################################################
class OrderedTable_View(Table_View):

    def get_items(self, resource, context):
        items = resource.handler.get_records_in_order()
        return list(items)


    def sort_and_batch(self, resource, context, items):
        # Sort
        sort_by = context.query['sort_by']
        if sort_by == 'order':
            reverse = context.query['reverse']
            ordered_ids = list(resource.handler.get_record_ids_in_order())
            f = lambda x: ordered_ids.index(x.id)
            items.sort(cmp=lambda x,y: cmp(f(x), f(y)), reverse=reverse)

            # Batch
            start = context.query['batch_start']
            size = context.query['batch_size']
            return items[start:start+size]

        return Table_View.sort_and_batch(self, resource, context, items)


    def get_table_columns(self, resource, context):
        columns = Table_View.get_table_columns(self, resource, context)
        columns.append(('order', MSG(u'Order')))

        return columns


    def get_item_value(self, resource, context, item, column):
        if column == 'order':
            ordered_ids = list(resource.handler.get_record_ids_in_order())
            return ordered_ids.index(item.id) + 1

        return Table_View.get_item_value(self, resource, context, item, column)


    table_actions = [RemoveButton, OrderUpButton, OrderDownButton,
                     OrderTopButton, OrderBottomButton]

    #######################################################################
    # Form Actions
    #######################################################################
    def action_remove(self, resource, context, form):
        ids = form['ids']
        for id in ids:
            resource.del_record(id)
        # Reindex the resource
        context.server.change_resource(resource)

        context.message = INFO(u'Record deleted.')


    def action_order_up(self, resource, context, form):
        ids = form['ids']
        if not ids:
            message = ERROR(u'Please select the resources to order up.')
            context.message = message
            return

        resource.handler.order_up(ids)
        context.message = INFO(u'Resources ordered up.')


    def action_order_down(self, resource, context, form):
        ids = form['ids']
        if not ids:
            message = ERROR(u'Please select the resources to order down.')
            context.message = message
            return

        resource.handler.order_down(ids)
        context.message = INFO(u'Resources ordered down.')


    def action_order_top(self, resource, context, form):
        ids = form['ids']
        if not ids:
            message = ERROR(u'Please select the resources to order on top.')
            context.message = message
            return

        resource.handler.order_top(ids)
        context.message = INFO(u'Resources ordered on top.')


    def action_order_bottom(self, resource, context, form):
        ids = form['ids']
        if not ids:
            message = ERROR(u'Please select the resources to order on bottom.')
            context.message = message
            return

        resource.handler.order_bottom(ids)
        context.message = INFO(u'Resources ordered on bottom.')




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from optparse import OptionParser
from sys import exit

# Import from itools
import itools
from itools.utils import become_daemon

# Import from ikaaro
from ikaaro.spool import Spool
from ikaaro.update import is_instance_up_to_date


def start(optios, target):
    # Check instance is up to date
    if not is_instance_up_to_date(target):
        print 'The instance is not up-to-date, please type:'
        print
        print '    $ icms-update.py <instance>'
        print
        exit(1)

    spool = Spool(target)
    pid = spool.get_pid()
    if pid is not None:
        print '[%s] The Mail Spool is already running.' % target
        exit(1)

    print '[%s] Start Mail Spool.' % target
    # Detach: redirect standard file descriptors to '/dev/null'
    if options.detach:
        become_daemon()

    # Start
    spool.start()


if __name__ == '__main__':
    # The command line parser
    usage = '%prog TARGET'
    version = 'itools %s' % itools.__version__
    description = ('Starts a spool server')
    parser = OptionParser(usage, version=version, description=description)
    parser.add_option(
        '-d', '--detach', action="store_true", default=False,
        help="Detach from the console.")

    options, args = parser.parse_args()
    if len(args) == 0:
        parser.error('The TARGET argument is missing.')

    # Start the spool
    start(options, args[0])



# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2008 Henry Obein <henry@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from datetime import datetime
from email.parser import HeaderParser
from os import fstat, getpid, remove as remove_file
from signal import signal, SIGINT
from smtplib import SMTP, SMTPRecipientsRefused, SMTPResponseException
from socket import gaierror
from time import sleep
from traceback import print_exc

# Import from itools
from itools.uri import get_absolute_reference2
from itools import vfs
from server import get_config
from utils import is_pid_running



class Spool(object):

    def __init__(self, target):
        target = get_absolute_reference2(target)
        self.target = target
        self.spool = target.resolve2('spool')
        self._stop = False

        # The SMTP host
        get_value = get_config(target).get_value
        self.smtp_host = get_value('smtp-host')
        self.smtp_login = get_value('smtp-login', default='').strip()
        self.smtp_password = get_value('smtp-password', default='').strip()

        # The logs
        self.activity_log_path = '%s/log/spool' % target.path
        self.activity_log = open(self.activity_log_path, 'a+')
        self.error_log_path = '%s/log/spool_error' % target.path
        self.error_log = open(self.error_log_path, 'a+')


    def get_pid(self):
        try:
            pid = open('%s/spool_pid' % self.target.path).read()
        except IOError:
            return None

        pid = int(pid)
        if is_pid_running(pid):
            return pid
        return None


    def start(self):
        # Pid
        open('%s/spool_pid' % self.target.path, 'w').write(str(getpid()))

        # Graceful stop
        signal(SIGINT, self.stop)

        # Go
        spool = vfs.open(self.spool)
        smtp_host = self.smtp_host
        log = self.log_activity
        while self._stop is False:
            sleep(10)
            # Find out emails to send
            locks = set()
            names = set()
            for name in spool.get_names():
                if name[-5:] == '.lock':
                    locks.add(name[:-5])
                else:
                    names.add(name)
            names.difference_update(locks)
            # Is there something to send?
            if len(names) == 0:
                continue

            # Open connection
            try:
                smtp = SMTP(smtp_host)
                if self.smtp_login and self.smtp_password:
                    smtp.login(self.smtp_login, self.smtp_password)
            except gaierror, excp:
                log('%s: "%s"' % (excp[1], smtp_host))
                continue
            except:
                self.log_error()
                continue
            # Send emails
            for name in names:
                try:
                    # Send message
                    message = spool.open(name).read()
                    headers = HeaderParser().parsestr(message)
                    subject = headers['subject']
                    from_addr = headers['from']
                    to_addr = headers['to']
                    # Send message
                    smtp.sendmail(from_addr, to_addr, message)
                    # Remove
                    spool.remove(name)
                    # Log
                    log('SENT "%s" from "%s" to "%s"' % (subject, from_addr,
                        to_addr))
                except (SMTPRecipientsRefused, SMTPResponseException):
                        # the SMTP server returns an error code
                        # or the recipient addresses has been refused
                        # Log
                        self.log_error()
                        # Remove
                        spool.remove(name)
                    # Other error ...
                except:
                    self.log_error()
            # Close connection
            smtp.quit()

        # Close files
        self.activity_log.close()
        self.error_log.close()
        # Remove pid file
        remove_file('%s/spool_pid' % self.target.path)


    def stop(self, n, frame):
        self._stop = True
        print 'Shutting down the mail spool (gracefully)...'


    def log_activity(self, msg):
        # The data to write
        data = '%s - %s\n' % (datetime.now(), msg)

        # Check the file has not been removed
        log = self.activity_log
        if fstat(log.fileno())[3] == 0:
            log = open(self.activity_log_path, 'a+')
            self.activity_log = log

        # Write
        log.write(data)
        log.flush()


    def log_error(self):
        # The data to write
        lines = [
            '\n',
            '%s\n' % ('*' * 78),
            'DATE: %s\n' % datetime.now(),
            '\n']
        data = ''.join(lines)

        # Check the file has not been removed
        log = self.error_log
        if fstat(log.fileno())[3] == 0:
            log = open(self.error_log_path, 'a+')
            self.error_log = log

        # Write
        log.write(data)
        print_exc(file=log) # FIXME Should be done before to reduce the risk
                            # of the log file being removed.
        log.flush()





# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2006-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007-2008 Nicolas Deram <nicolas@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from email.charset import add_charset, add_codec, QP
from email.MIMEText import MIMEText
from email.MIMEMultipart import MIMEMultipart
from email.Utils import formatdate
from email.header import Header
import traceback

# Import from itools
from itools import get_abspath
from itools.gettext import MSG
from itools.handlers import ConfigFile
from itools.stl import stl
from itools.uri import Path
from itools.web import get_context, STLView

# Import from ikaaro
from folder import Folder
from registry import register_resource_class, get_resource_class
from skins import UI, ui_path
from user import UserFolder
from utils import crypt_password
from website import WebSite


# itools source and target languages
config = get_abspath('setup.conf')
config = ConfigFile(config)
itools_source_language = config.get_value('source_language')
itools_target_languages = config.get_value('target_languages')


# Force email to send UTF-8 mails in plain text
add_charset('utf-8', QP, None, 'utf-8')
add_codec('utf-8', 'utf_8')



class NotFoundView(STLView):
    template = '/ui/root/not_found.xml'

    def get_namespace(self, resource, context):
        return {'uri': str(context.uri)}


class ForbiddenView(STLView):
    template = '/ui/root/forbidden.xml'

    def POST(self, resource, context):
        return self.GET




class Root(WebSite):

    class_id = 'iKaaro'
    class_version = '20071215'
    class_title = MSG(u'iKaaro')
    class_icon16 = 'icons/16x16/root.png'
    class_icon48 = 'icons/48x48/root.png'
    class_control_panel = ['browse_users', 'add_user', 'edit_virtual_hosts',
                           'edit_security_policy', 'edit_languages',
                           'edit_contact_options']


    __fixed_handlers__ = ['users', 'ui']


    __roles__ = [
        {'name': 'admins', 'title': MSG(u'Admin')}]


    @staticmethod
    def _make_resource(cls, folder, email, password):
        # The metadata
        metadata = cls.build_metadata(admins=('0',))
        folder.set_handler('.metadata', metadata)
        # User Folder
        users = UserFolder.build_metadata(title={'en': u'Users'})
        folder.set_handler('users.metadata', users)
        # Default User
        password = crypt_password(password)
        user_class = get_resource_class('user')
        user = user_class.build_metadata(email=email, password=password)
        folder.set_handler('users/0.metadata', user)
        # Return
        return cls(metadata)


    @classmethod
    def get_metadata_schema(cls):
        schema = WebSite.get_metadata_schema()
        return schema


    ########################################################################
    # Override itools.web.root.Root
    ########################################################################
    def get_user(self, name):
        users = self.get_resource('users')
        if users.has_resource(name):
            return users.get_resource(name)
        return None


    def get_user_from_login(self, username):
        """Return the user identified by its unique e-mail or username, or
        return None.
        """
        # Search the user by username (login name)
        results = self.search(username=username)
        if len(results) == 0:
            return None
        # Get the user
        brain = results.get_documents()[0]
        return self.get_user(brain.name)


    ########################################################################
    # Publish
    ########################################################################
    forbidden = ForbiddenView()


    def internal_server_error(self, context):
        namespace = {'traceback': traceback.format_exc()}

        handler = self.get_resource('/ui/root/internal_server_error.xml')
        return stl(handler, namespace, mode='html')


    ########################################################################
    # Traverse
    ########################################################################
    def _get_resource(self, name):
        if name == 'ui':
            ui = UI(ui_path)
            ui.database = self.metadata.database
            return ui
        return Folder._get_resource(self, name)


    def _get_names(self):
        names = [ x for x in Folder._get_names(self) if x ]
        return names + ['ui']


    ########################################################################
    # API
    ########################################################################
    def get_document_types(self):
        return WebSite.get_document_types(self) + [WebSite]


    ########################################################################
    # Search
    def search(self, query=None, **kw):
        catalog = get_context().server.catalog
        return catalog.search(query, **kw)


    ########################################################################
    # Skins
    def get_skin(self):
        context = get_context()
        # Back-Office
        hostname = context.uri.authority.host
        if hostname[:3] in ['bo.', 'bo-']:
            return self.get_resource('ui/aruni')
        # Fron-Office
        skin = context.site_root.class_skin
        return self.get_resource(skin)


    def get_available_languages(self):
        """Returns the language codes for the user interface.
        """
        source = itools_source_language
        target = itools_target_languages
        # A package based on itools
        cls = self.__class__
        if cls is not Root:
            exec('import %s as pkg' % cls.__module__.split('.', 1)[0])
            config = Path(pkg.__path__[0]).resolve2('setup.conf')
            config = ConfigFile(str(config))
            source = config.get_value('source_language', default=source)
            target = config.get_value('target_languages', default=target)

        target = target.split()
        if source in target:
            target.remove(source)

        target.insert(0, source)
        return target


    ########################################################################
    # Email
    def send_email(self, to_addr, subject, from_addr=None, text=None,
                   html=None, encoding='utf-8', subject_with_host=True,
                   return_receipt=False):
        # Check input data
        if not isinstance(subject, unicode):
            raise TypeError, 'the subject must be a Unicode string'
        if len(subject.splitlines()) > 1:
            raise ValueError, 'the subject cannot have more than one line'
        if text and not isinstance(text, unicode):
            raise TypeError, 'the text must be a Unicode string'
        if html and not isinstance(html, unicode):
            raise TypeError, 'the html must be a Unicode string'

        # Figure out the from address
        context = get_context()
        server = context.server
        if from_addr is None:
            user = context.user
            if user is not None:
                from_addr = user.get_property('email')
            if not from_addr:
                from_addr = server.smtp_from

        # Set the subject
        subject = subject.encode(encoding)
        if subject_with_host is True:
            host = context.uri.authority.host
            subject = '[%s] %s' % (host, subject)
        # Build the message
        message = MIMEMultipart('related')
        message['Subject'] = Header(subject, encoding)
        message['Date'] = formatdate(localtime=True)
        message['From'] = from_addr
        if isinstance(to_addr, tuple):
            real_name, address = to_addr
            to_addr = '%s <%s>' % (Header(real_name, encoding), address)
        message['To'] = to_addr
        # Return Receipt
        if return_receipt is True:
            # Somewhat standard
            message['Disposition-Notification-To'] = from_addr
            # XXX For Outlook 2000
            message['Return-Receipt-To'] = from_addr
        # Create MIMEText
        if html:
            html = html.encode(encoding)
            message_html = MIMEText(html, 'html', _charset=encoding)
        if text:
            text = text.encode(encoding)
            message_text = MIMEText(text, _charset=encoding)
        # Attach MIMETEXT to message
        if text and html:
            message_alternative = MIMEMultipart('alternative')
            message.attach(message_alternative)
            message_alternative.attach(message_text)
            message_alternative.attach(message_html)
        elif html:
            message.attach(message_html)
        elif text:
            message.attach(message_text)
        # Send email
        server.send_email(message)


    not_found = NotFoundView()


###########################################################################
# Register
###########################################################################
register_resource_class(Root)
